/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.io.IOException;

import com.aquenos.epics.jackie.client.DefaultClientThreadingStrategy;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.SimpleCommunicationController;

/**
 * <p>
 * Thread that acts as a {@link CommunicationController}. This implementation
 * uses a {@link SimpleCommunicationController} that is driven by its own
 * thread. The thread created is a daemon thread, so it does not prevent the JVM
 * from being shutdown.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultClientThreadingStrategy}
 * .
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class CommunicationThread extends Thread {

    private volatile boolean destroyRequested = false;
    private ErrorHandler errorHandler;
    private SimpleCommunicationController communicationController;

    /**
     * Creates a communication thread. The thread uses the specified name and is
     * created as a daemon thread. However, the thread is not started
     * automatically but has to be started by calling its {@link #start()}
     * method.
     * 
     * @param threadName
     *            name of the new thread.
     * @param errorHandler
     *            error handler that is used to log errors that are thrown by
     *            the communication controller.
     */
    public CommunicationThread(String threadName, ErrorHandler errorHandler) {
        super(threadName);
        assert (threadName != null);
        assert (errorHandler != null);
        try {
            this.communicationController = new SimpleCommunicationController();
        } catch (IOException e) {
            // If we cannot even create the controller, there is no reasonable
            // way how we could continue. On the other hand, this is very
            // unlikely because it only happens if no selector can be opened.
            throw new RuntimeException(
                    "Could not create a communication controller.", e);
        }
        this.errorHandler = errorHandler;
        // We do not want communication threads to stop the VM from being
        // shutdown. Therefore we mark them as daemon threads.
        this.setDaemon(true);
    }

    @Override
    public void run() {
        while (!destroyRequested && !Thread.currentThread().isInterrupted()) {
            try {
                communicationController.processTimersAndIO();
            } catch (IOException e) {
                errorHandler.handleError(DefaultClientThreadingStrategy.class,
                        e, "Exception while trying to process I/O and timers.");
                // An I/O exception might be caused by the select call,
                // therefore we sleep a bit to avoid spinning the CPU at 100% if
                // the select call fails continuously.
                try {
                    Thread.sleep(10);
                } catch (InterruptedException e1) {
                    Thread.currentThread().interrupt();
                }
            } catch (Throwable e) {
                errorHandler.handleError(DefaultClientThreadingStrategy.class,
                        e, "Exception while trying to process I/O and timers.");
            }
        }
    }

    /**
     * Stops the communication thread gracefully. It marks the thread as
     * destroyed and wakes it up if it is currently waiting for a channel to
     * become available for I/O. It also interrupts the thread in order to make
     * sure it is stopped even if it is sleeping for a different reason than
     * waiting on the selector.
     */
    public void stopGracefully() {
        destroyRequested = true;
        communicationController.wakeUp();
        // Typically, setting destroyRequested to true and waking up the
        // controller should be enough. However, if the thread waits for some
        // operation (it should not because it is designed to work
        // asynchronously), interrupting the thread should make sure that the
        // the thread actually quits. We do not try to stop the thread the hard
        // way because this can cause more harm than good.
        this.interrupt();
    }

    /**
     * Returns the communication controller that is managed by this thread. The
     * communication controller returned can be assigned to communication
     * processors that should be processed by this thread.
     * 
     * @return communication controller for this thread.
     */
    public CommunicationController getCommunicationController() {
        return communicationController;
    }

}
