/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.IllegalBlockingModeException;
import java.nio.channels.IllegalSelectorException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;

/**
 * <p>
 * Interface for a component that manages I/O activities. A communication
 * controller typically controls one or more {@link ChannelProcessor}s and
 * {@link TimerProcessor}s calling one of their process methods at the right
 * time.
 * </p>
 * 
 * <p>
 * A communication controller allows communication processors to register for
 * I/O events for channels and for timer events. The controller will call the
 * respective processor's {@link ChannelProcessor#processIO() processIO()} or
 * {@link TimerProcessor#processTimer()} method each time such an event occurs.
 * This way, one thread can handle I/O and timer events for multiple components.
 * </p>
 * 
 * <p>
 * A communication controller must be thread safe, meaning that a communication
 * processor must be able to call the controller's methods at any time, not just
 * from one of the methods called by the controller.
 * </p>
 * 
 * <p>
 * On the other hand, a communication controller may never call more than one
 * processor's <code>process</code> method in parallel. This is important,
 * because processor's might not be thread-safe and their might be hidden
 * dependencies between processors registered with the same communication
 * controller.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface CommunicationController {

    /**
     * <p>
     * Registers a channel with the communication controller. This tells the
     * communication controller that the channel processor is interested in I/O
     * events for this channel. The channel processor can exactly define which
     * events it is interested in through the {@link SelectionKey} returned by
     * this method.
     * </p>
     * 
     * <p>
     * When this communication controller detects that one of the I/O events the
     * <code>channelProcessor</code> has registered for has occurred, it calls
     * the <code>channelProcessor</code>'s {@link ChannelProcessor#processIO()
     * processIO()} method.
     * </p>
     * 
     * <p>
     * The selection key returned by this method may be a wrapper that only
     * provides a subset of the operations defined in the {@link SelectionKey}
     * interface. In particular, its {@link SelectionKey#selector() selector()}
     * methods may throw an {@link UnsupportedOperationException}. This should
     * not be a problem because a processor does not have good reasons to call
     * this method anyway.
     * </p>
     * 
     * <p>
     * A typical implementation of this method will call the
     * <code>channel</code>'s
     * {@link SelectableChannel#register(java.nio.channels.Selector, int, Object)}
     * method and will throw any exception thrown by this method.
     * </p>
     * 
     * @param channelProcessor
     *            the channel processor that registers the channel. Each channel
     *            may only be registered for a single channel processor.
     * @param channel
     *            the channel that shall be registered. Each channel may only be
     *            registered for a single channel processor.
     * @return the selection key representing the channel's registration with a
     *         {@link Selector}. This selection key might not be the original
     *         selection key but a wrapper that does not implement the
     *         <code>selector()</code> and <code>attach(Object)</code> methods.
     * @throws ClosedChannelException
     *             if the <code>channel</code> channel is closed.
     * @throws IllegalBlockingModeException
     *             if the <code>channel</code> is in blocking mode.
     * @throws IllegalSelectorException
     *             if the <code>channel</code> was not created by the same
     *             provider as this communication controller's selector.
     * @throws CancelledKeyException
     *             if the <code>channel</code> is currently registered with this
     *             communication controller's selector but the corresponding key
     *             has already been cancelled.
     */
    SelectionKey registerChannel(ChannelProcessor channelProcessor,
            SelectableChannel channel) throws ClosedChannelException;

    /**
     * <p>
     * Registers for a timer event. When the time specified by
     * <code>eventTime</code> has come or passed, this communication controller
     * calls the processor's {@link TimerProcessor#processTimer()
     * processTimer()} method.
     * </p>
     * 
     * <p>
     * Multiple timer registrations for the same <code>timerProcessor</code> may
     * result in a single call to <code>processTimer()</code> if the
     * <code>eventTime</code> of all of these registrations has passed.
     * </p>
     * 
     * <p>
     * The general contract is that a processor's <code>processTimer()</code>
     * method may be called an indefinite amount of time after, but never before
     * the <code>eventTime</code> has passed.
     * </p>
     * 
     * @param timerProcessor
     *            the processor that is registering for a timer event.
     * @param eventTime
     *            Java time-stamp (as returned by
     *            {@link System#currentTimeMillis()}) that specifies the point
     *            in time when the timer event should occur. The processor's
     *            process method may be called later, but never before this
     *            time.
     */
    void registerTimer(TimerProcessor timerProcessor, long eventTime);

    /**
     * Tells whether the current thread is the communication thread. The
     * communication thread is the thread from which a communication controller
     * calls the process methods of {@link ChannelProcessor}s and
     * {@link TimerProcessor}s. This method can be used by processors that are
     * not thread safe to detect whether code is run in the correct thread.
     * 
     * @return <code>true</code> if the current thread is the communication
     *         thread, <code>false</code> otherwise.
     */
    boolean inCommunicationThread();

}
