/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.exception.SerializedPayloadTooLargeException;
import com.aquenos.epics.jackie.common.exception.ShortPayloadException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus.StatusMessage;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueCodec;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Representation of a <code>CA_PROTO_EVENT_ADD</code> message that is sent from
 * a server to a client. This message is sent each time there is an event that
 * the client subscribed to. If the event represents a successful read, the
 * message contains the channel's current value. If the event represents a read
 * failure, the message contains the error code and the value is undefined (
 * <code>null</code>). A server should sent the first subscription event right
 * after registering a subscription, so that the client has an up-to-date value
 * immediately.
 * </p>
 * 
 * <p>
 * This message can easily be confused with a
 * {@link ChannelAccessCancelSubscriptionServerMessage}. The only way to tell
 * the two apart is by their payload size. This message always has a non-zero
 * payload size while a {@link ChannelAccessCancelSubscriptionServerMessage}
 * always has a zero payload size.
 * </p>
 * 
 * @see ChannelAccessCommand#CA_PROTO_EVENT_ADD
 * @see ChannelAccessSubscriptionClientMessage
 * @see ChannelAccessCancelSubscriptionServerMessage
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessSubscriptionServerMessage extends
        ChannelAccessSubscriptionMessage {

    private static byte[] PADDING_BYTES = new byte[ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT];

    private short dataType;
    private int count;
    private ChannelAccessStatus status;
    private int subscriptionId;
    private ChannelAccessGettableValue<?> value;
    private byte[] rawValue;

    /**
     * Creates a <code>CA_PROTO_EVENT_ADD</code> message. This constructor
     * should be used to create a message that is sent from a server to a client
     * in response to a {@link ChannelAccessSubscriptionClientMessage} sent by
     * the client.
     * 
     * @param dataType
     *            data type of the value in the message's payload (never
     *            <code>null</code>). This should always be the type that the
     *            client requested. If the type requested by the client is not
     *            supported by the server, an error message should be sent
     *            instead of this message.
     * @param count
     *            number of value elements in the message's payload. This should
     *            always be the number of elements that the client requested. If
     *            the number requested by the client is not supported by the
     *            server, an error message should be sent instead of this
     *            message. Must not be negative.
     * @param status
     *            status code for the subscription event (never
     *            <code>null</code>). If the event contains a valid value, the
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL
     *            ECA_NORMAL} code should be sent and <code>value</code> should
     *            not be <code>null</code>. If there is no valid value because
     *            an error occurred, the respective error code should be sent
     *            and <code>value</code> should be <code>null</code>.
     * @param subscriptionId
     *            client-assigned subscription ID. This ID must be the same ID
     *            that was sent with the client's subscription request.
     * @param value
     *            current value of the channel or <code>null</code> if this
     *            message represent an error (<code>status</code> is not
     *            <code>ECA_NORMAL</code>).
     * @throws IllegalArgumentException
     *             if the <code>status</code> is <code>ECA_NORMAL</code> and the
     *             <code>value</code> is <code>null</code>, if the
     *             <code>status</code> is not <code>ECA_NORMAL</code> and the
     *             <code>value</code> is not <code>null</code>, if the
     *             <code>dataType</code> does not match the <code>value</code>'s
     *             data type, and if <code>count</code> is negative or greater
     *             than the <code>value</code>'s number of elements.
     */
    public ChannelAccessSubscriptionServerMessage(
            ChannelAccessValueType dataType, int count,
            ChannelAccessStatus status, int subscriptionId,
            ChannelAccessGettableValue<?> value) {
        super();

        if (count < 0) {
            throw new IllegalArgumentException(
                    "The count must not be negative.");
        }

        boolean successMessage = status
                .matchesMessage(StatusMessage.ECA_NORMAL);
        if (value == null) {
            if (successMessage) {
                throw new IllegalArgumentException(
                        "The value must not be null if the status is ECA_NORMAL.");
            }
        } else {
            if (!successMessage) {
                throw new IllegalArgumentException(
                        "The value must be null if the status is not ECA_NORMAL.");
            }
            if (!dataType.equals(value.getType())) {
                throw new IllegalArgumentException("The specified data type ("
                        + dataType.toString()
                        + ") does not match the value's data type ("
                        + value.getType() + ").");
            }
            if (count > value.getValueSize()) {
                throw new IllegalArgumentException("The count (" + count
                        + ") exceeds the number of elements in the value ("
                        + value.getValueSize() + ").");
            }
        }

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.status = status;
        this.subscriptionId = subscriptionId;
        // Make sure that we only hand out a read-only version of the value.
        // As we never write to the value, we can simply replace the reference.
        if (value != null) {
            this.value = value.asReadOnlyValue();
        }
    }

    private ChannelAccessSubscriptionServerMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, ChannelAccessStatus status, int subscriptionId,
            ChannelAccessGettableValue<?> value, byte[] rawValue) {
        super(messageHeader);
        this.dataType = dataType;
        this.count = count;
        this.status = status;
        this.subscriptionId = subscriptionId;
        if (value != null) {
            this.value = value.asReadOnlyValue();
        }
        this.rawValue = rawValue;
    }

    /**
     * Returns the data type of the value in this message's payload. This should
     * always match the data type that the client requested. Returns
     * <code>null</code> if the specified data-type is unknown to this
     * implementation.
     * 
     * @return data-type of the value or <code>null</code> if the numeric
     *         identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            ChannelAccessValueType type = ChannelAccessValueType
                    .forTypeNumber(this.dataType);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (type.isGettable()) {
                return type;
            } else {
                return null;
            }
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements the value in this message's payload has.
     * This should always match the element count that the client requested.
     * 
     * @return number of value elements.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the status of the subscription event. If the subscription event
     * contains a valid value, the status should be
     * {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL} and
     * {@link #getValue()} should return a non-null value. If the event
     * indicates a problem, the status should be set to the appropriate error
     * code and {@link #getValue()} should return <code>null</code>.
     * 
     * @return status code of the event. Should be <code>ECA_NORMAL</code> for a
     *         success event and the error code for an error event.
     */
    public ChannelAccessStatus getStatus() {
        return status;
    }

    /**
     * Returns the client-assigned subscription ID. The subscription ID is used
     * by the client in order to map events sent in response to a subscription
     * to the respective subscription. This must always be the ID that was sent
     * by the client. This number might be negative or zero.
     * 
     * @return client-assigned subscription identifier.
     */
    public int getSubscriptionId() {
        return subscriptionId;
    }

    /**
     * Returns the current value of the channel. If the event indicates an
     * error, this method returns <code>null</code> and {@link #getStatus()}
     * returns the corresponding status code. This method also returns
     * <code>null</code> if the data-type sent by the server is not supported by
     * this implementation ({@link #getDataType()} returns <code>null</code>).
     * 
     * @return current value of the channel or <code>null</code> if the event
     *         indicates an error and thus there is no value. The value's data
     *         type is the same that is returned by {@link #getDataType()} and
     *         the value's element count is the number returned by
     *         {@link #getCount()}.
     */
    public ChannelAccessGettableValue<?> getValue() {
        if (value == null) {
            return null;
        } else {
            return value;
        }
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (rawValue == null) {
            // This is the usual case. The data-type is valid and we either have
            // a value or this is an error message.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            serializeHeader(byteSink, payloadSize + paddingSize, dataType,
                    count, status.toStatusCode(), subscriptionId, version,
                    maxPayloadSize);
            if (value != null) {
                ChannelAccessValueCodec.encodeGettableValue(byteSink, value,
                        count);
                byteSink.putByteArray(PADDING_BYTES, 0, paddingSize);
            } else {
                // If we have no value and no raw value, this must be an error
                // message, so we send an empty payload. We can add a whole
                // block of data at once because we know that the total size
                // must be aligned to the block size.
                int numberOfBlocks = (payloadSize + paddingSize)
                        / ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT;
                for (int i = 0; i < numberOfBlocks; i++) {
                    byteSink.putByteArray(PADDING_BYTES);
                }

            }
        } else {
            // We have a raw value. This means that we could not handle the type
            // and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            serializeHeader(byteSink, rawValue.length, dataType, count,
                    status.toStatusCode(), subscriptionId, version,
                    maxPayloadSize);
            byteSink.putByteArray(rawValue);
        }
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        // maxPayloadSize should be aligned.
        assert (maxPayloadSize % ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT == 0);
        ChannelAccessValueType dataTypeEnum = ChannelAccessValueType
                .forTypeNumber(dataType);
        if (rawValue == null) {
            // This is the usual case. The data-type is valid and we either have
            // a value or this is an error message.
            int maxCount = ChannelAccessValueCodec.calculateMaxCount(
                    dataTypeEnum, maxPayloadSize);
            if (count > maxCount || count < 0) {
                throw new SerializedPayloadTooLargeException(
                        "Cannot send message with value type "
                                + dataTypeEnum.toString() + " and count "
                                + (count & 0xffffffffL)
                                + " because payload size limit is set to "
                                + maxPayloadSize + ".");
            }
            int payloadSize = ChannelAccessValueCodec.calculatePayloadSize(
                    ChannelAccessValueType.forTypeNumber(dataType), count);
            int paddingSize = calculatePaddingSize(payloadSize);
            verifyHeader(payloadSize + paddingSize, count, version,
                    maxPayloadSize);
        } else {
            // We have a raw value. This means that we could not handle the type
            // and thus we could not deserialize the value. However, we can
            // simply use the raw value.
            verifyHeader(rawValue.length, count, version, maxPayloadSize);
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessSubscriptionServerMessage other = (ChannelAccessSubscriptionServerMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count)
                .append(this.status, other.status)
                .append(this.subscriptionId, other.subscriptionId)
                .append(this.value, other.value)
                .append(this.rawValue, other.rawValue).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(status)
                .append(subscriptionId).append(value).append(rawValue)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", status=");
        sb.append(status);
        sb.append(", subscriptionId=");
        sb.append(subscriptionId);
        if (value != null) {
            sb.append(", value=");
            sb.append(value);
        }
        if (rawValue != null) {
            sb.append(", rawValue=");
            addByteArrayToStringBuilder(sb, rawValue);
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param charset
     *            encoding to use for deserializing the host name. Usually, host
     *            names should consist of ASCII characters only, however this is
     *            not strictly enforced by the protocol, so the encoding might
     *            matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws ShortPayloadException
     *             if the message's payload size is less than the size needed
     *             for the data type and count.
     */
    protected static ChannelAccessSubscriptionServerMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Charset charset) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_EVENT_ADD
                .getCommandNumber());

        short dataTypeNumber = messageHeader.getDataType();
        ChannelAccessValueType dataType;
        try {
            dataType = ChannelAccessValueType.forTypeNumber(dataTypeNumber);
            // We treat a data type that is not valid for this message the same
            // way we treat an unknown data type.
            if (!dataType.isGettable()) {
                dataType = null;
            }
        } catch (IllegalArgumentException e) {
            dataType = null;
        }
        int count = messageHeader.getCount();
        ChannelAccessStatus status = ChannelAccessStatus
                .forStatusCode(messageHeader.getCID());
        int subscriptionId = messageHeader.getContextSpecific();
        int actualPayloadSize = messageHeader.getPayloadSize();
        // If the payload size is zero, a
        // ChannelAccessCancelSubscriptionServerMessage should be created
        // instead.
        assert (actualPayloadSize != 0);
        ChannelAccessGettableValue<?> value;
        byte[] rawValue;
        if (dataType == null) {
            value = null;
            rawValue = byteSource.getByteArray(actualPayloadSize);
        } else {
            if (status.matchesMessage(StatusMessage.ECA_NORMAL)) {
                int expectedPayloadSize;
                try {
                    // calculatePayloadSize throws an IllegalArgumentException
                    // if count is negative or too large to represent the
                    // resulting number as an int.
                    expectedPayloadSize = ChannelAccessValueCodec
                            .calculatePayloadSize(dataType, count);
                } catch (IllegalArgumentException e) {
                    throw new ShortPayloadException("Payload size of "
                            + actualPayloadSize + " is too small to hold a "
                            + dataType.toString() + " with "
                            + (count & 0xffffffffL) + " elements.");
                }
                if (actualPayloadSize < expectedPayloadSize) {
                    throw new ShortPayloadException(
                            "Received a CA_PROTO_EVENT_ADD message with a payload size of "
                                    + actualPayloadSize
                                    + ". However, the message should have a payload of at least "
                                    + expectedPayloadSize + " bytes.");
                }
                value = ChannelAccessValueCodec.decodeGettableValue(byteSource,
                        dataType, expectedPayloadSize, count, charset);
                byteSource.skip(actualPayloadSize - expectedPayloadSize);
            } else {
                byteSource.skip(actualPayloadSize);
                value = null;
            }
            rawValue = null;
        }
        return new ChannelAccessSubscriptionServerMessage(messageHeader,
                dataTypeNumber, count, status, subscriptionId, value, rawValue);
    }

}
