/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;

/**
 * Representation of a <code>CA_PROTO_VERSION</code> message that is sent from a
 * client to a server or the other way round via UDP. This message is usually
 * send at the start of each UDP packet in order to communicate the protocol
 * version supported by the client (or server) and in order to communicate the
 * sequence number associated with the search requests in the same packet.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_VERSION
 * @see ChannelAccessVersionTCPClientMessage
 * @see ChannelAccessVersionTCPServerMessage
 */
public final class ChannelAccessVersionUDPMessage extends
        ChannelAccessVersionMessage {

    private boolean sequenceNumberIsValid;
    private int sequenceNumber;

    /**
     * Creates a <code>CA_PROTO_VERSION</code> message. This constructor should
     * be used to create a message that is supposed to be sent in any direction
     * using the UDP protocol. It should be the first message in each UDP
     * packet.
     * 
     * @param version
     *            Channel Access version supported by the sender.
     * @param sequenceNumberIsValid
     *            indicator whether the sequence number should be considered to
     *            be valid. If this flag is set, the sequence number might be
     *            used to calculate an estimation of the round-trip time.
     * @param sequenceNumber
     *            sequence number. This number is supposed to be incremented
     *            with each packet sent. It might be used to calculate an
     *            estimation of the round-trip time.
     */
    public ChannelAccessVersionUDPMessage(ChannelAccessVersion version,
            boolean sequenceNumberIsValid, int sequenceNumber) {
        super(version);

        if (version == null) {
            throw new NullPointerException(
                    "Channel Acccess version must not be null.");
        }

        this.sequenceNumberIsValid = sequenceNumberIsValid;
        this.sequenceNumber = sequenceNumber;
    }

    private ChannelAccessVersionUDPMessage(
            ChannelAccessMessageHeader messageHeader,
            ChannelAccessVersion version, boolean sequenceNumberIsValid,
            int sequenceNumber) {
        super(messageHeader, version);

        this.sequenceNumberIsValid = sequenceNumberIsValid;
        this.sequenceNumber = sequenceNumber;
    }

    /**
     * Returns the sequence-number-is-valid flag. This flag may be set by a
     * client when sending a message over UDP and should be returned by the
     * server together with the respective sequence number.
     * 
     * @return sequence-number-is-valid flag, indicating whether the number
     *         returned by {@link #getSequenceNumber()} should be used to update
     *         the round-trip time estimate.
     */
    public boolean isSequenceNumberValid() {
        return sequenceNumberIsValid;
    }

    /**
     * Returns the sequence number of this message. This number may be used to
     * update the round-trip time estimate, but only if
     * {@link #isSequenceNumberValid()} returns <code>true</code>.
     * 
     * @return sequence number. The sequence number should be increased with
     *         each packet sent but may be negative if it wraps.
     */
    public int getSequenceNumber() {
        return sequenceNumber;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, (short) (sequenceNumberIsValid ? 1 : 0),
                this.version.getMinorVersionAsShort(), sequenceNumber, 0,
                version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessVersionUDPMessage other = (ChannelAccessVersionUDPMessage) obj;
        return new EqualsBuilder()
                .append(this.sequenceNumberIsValid, other.sequenceNumberIsValid)
                .append(this.sequenceNumber, other.sequenceNumber).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(sequenceNumberIsValid).append(sequenceNumber)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("version=");
        sb.append(version);
        sb.append(", sequenceNumberIsValid=");
        sb.append(sequenceNumberIsValid);
        sb.append(", sequenceNumber=");
        sb.append(sequenceNumber);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessVersionUDPMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_VERSION
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessVersionUDPMessage(messageHeader,
                ChannelAccessVersion.forMinorVersion(messageHeader.getCount()),
                (messageHeader.getDataType() & 1) != 0, messageHeader.getCID());
    }

}
