/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.atomic.AtomicInteger;

/**
 * <p>
 * Pool of integer IDs. The pool is both thread safe and lock free. IDs are
 * acquired and released as primitive integer numbers, so the caller is
 * responsible for making sure that it does not use an ID after it has been
 * released and that it does not release an ID that has not been acquired or has
 * already been released. For a more self-contained variant see
 * {@link IntegerIdPool}.
 * </p>
 * 
 * <p>
 * The pool generates integer IDs from a configurable range, ensuring that each
 * ID is only handed out once. In order to accommodate for applications that
 * need a large number of IDs over their life-cycle, the pool supports the reuse
 * of IDs if they are released when they are not needed any longer.
 * </p>
 * 
 * <p>
 * Optionally, the pool can be configured to block released IDs from immediate
 * reuse for a certain time. This is useful in environments where IDs leave the
 * scope of the application and thus external entities might have dangling
 * references to those IDs. In this case, configuring a time for which the ID is
 * blocked from reuse after having been released can help in avoiding a
 * situation in which an external entity uses an ID that has now been reused for
 * a different purpose.
 * </p>
 * 
 * <p>
 * Internally, this implementation uses a {@link SimpleConcurrentQueue} for
 * keeping a list of IDs that have been released and can be reused. If this list
 * is empty, a completely new ID is acquired by incrementing an atomic counter
 * that keeps track of the IDs that have never been used.
 * </p>
 * 
 * <p>
 * This implementation is more efficient than keeping a list of IDs in use and
 * checking incrementally growing IDs against this list. While such an algorithm
 * would have a worst-case performance of O(n) for acquire operations, this
 * implementation has O(1) performance for both acquire and release operations.
 * This advantage comes at the cost of a significant memory footprint if a large
 * number of IDs is acquired, subsequently released, and no IDs are acquired
 * subsequently. In this case, this pool uses memory proportional to the number
 * of IDs that have been in use in parallel.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class BasicIntegerIdPool {

    /**
     * Simple class holding a reusable ID and the time-stamp when it can be
     * reused.
     */
    private static class ReusableId {

        /**
         * ID.
         */
        public int id;

        /**
         * Time after which the ID may be reused (as a Java time-stamp).
         */
        public long reuseAfter;

    }

    /**
     * Smallest ID that is assigned.
     */
    private int minimumId;

    /**
     * Largest ID that is assigned.
     */
    private int maximumId;

    /**
     * Is {@link #minimumId} greater than {@link #maximumId}?
     */
    private boolean wrapAround;

    /**
     * Time (in milliseconds) for which an ID is blocked from being reused after
     * having been released.
     */
    private long blockFromReuseTime;

    /**
     * May IDs that have been released but are still blocked from reuse be
     * reused if otherwise the pool is exhausted?
     */
    private boolean earlyReuseOnExhaustion;

    /**
     * Counter that keeps the next ID that has never been assigned.
     */
    private AtomicInteger nextId;

    /**
     * Queue holding reusable IDs. These are IDs that have been returned to the
     * pool by calls to {@link #releaseId(int)}. The IDs are roughly ordered by
     * their reuse time because this time is calculated based on the current
     * time when they are added. Due to concurrency issues, they might be
     * slightly out of order, but only to a degree that is acceptable for this
     * application.
     */
    private SimpleConcurrentQueue<ReusableId> reusableIds;

    /**
     * <p>
     * Creates a new pool for integer IDs. All IDs returned by the pools
     * {@link #acquireId()} method are guaranteed to be in the range from
     * <code>minimumId</code> to <code>maximumId</code> (both inclusive).
     * </p>
     * 
     * <p>
     * If <code>minimumId</code> is greater than <code>maximumId</code>, the
     * pool will start at <code>minimumId</code> counting upwards until
     * {@link Integer#MAX_VALUE} is reached. Then, the counter will wrap around
     * continuing with negative numbers until <code>maximumId</code> is reached.
     * </p>
     * 
     * <p>
     * <code>minimumId</code> and <code>maximumId</code> must be chosen so that
     * there is at least one integer number that is not covered by the range.
     * This number is needed because it will be internally used as a marker when
     * the counter has been exhausted.
     * </p>
     * 
     * <p>
     * The <code>blockFromReuseTime</code> parameter specifies the amount of
     * time in milliseconds that must pass between an ID being returned by
     * calling {@link #releaseId(int)} and {@link #acquireId()} returning the
     * same ID. This is useful for applications where an external entity might
     * keep an ID for some time and it is important to distinguish the now
     * invalid ID from newly acquired IDs.
     * </p>
     * 
     * <p>
     * If <code>earlyReuseOnExhaustion</code> is <code>true</code>,
     * {@link #acquireId()} will not fail immediately if no ID can be reused yet
     * and no new ID can be assigned because all possible IDs are already in
     * use. Instead of that, it will try to find an ID that has been released,
     * but for which the <code>blockFromReuseTime</code> has not passed yet.
     * This can be used as a last resort, when it is more important to make sure
     * that IDs can be acquired than it is important to ensure that released IDs
     * are not reused for a certain time.
     * </p>
     * 
     * @param minimumId
     *            minimum number used as an ID (inclusive). If
     *            <code>minimumId</code> is greater than <code>maximumId</code>,
     *            IDs that are less than <code>minimumId</code> are used if they
     *            are also less than or equal to <code>maximumId</code>.
     * @param maximumId
     *            maximum number used as an ID (inclusive). If
     *            <code>maximumId</code> is less than <code>minimumId</code>,
     *            IDs that are greater than <code>maximumId</code> are used if
     *            they are also greater than or equal to <code>minimumId</code>.
     * @param blockFromReuseTime
     *            number of milliseconds to wait before reusing an ID after it
     *            has been released. A value of <code>0</code> means that IDs
     *            can be reused immediately.
     * @param earlyReuseOnExhaustion
     *            if <code>true</code> IDs which have been released less than
     *            <code>blockFromReuseTime</code> milliseconds ago are still
     *            used if otherwise the pool would be exhausted.
     * @throws IllegalArgumentException
     *             if <code>minimumId</code> and <code>maximumId</code> have not
     *             been chosen so that there is at least one unusable ID or if
     *             blockFromReuseTime is negative.
     */
    public BasicIntegerIdPool(int minimumId, int maximumId,
            long blockFromReuseTime, boolean earlyReuseOnExhaustion) {
        this.minimumId = minimumId;
        this.maximumId = maximumId;
        // We need a gap of at least one ID between the maximum and minimum ID.
        // This is important because this is the only way to detect when our
        // counter for new IDs has been exhausted.
        if (minimumId == maximumId || maximumId == minimumId - 1
                || minimumId == Integer.MIN_VALUE
                && maximumId == Integer.MAX_VALUE) {
            throw new IllegalArgumentException(
                    "The minimum ID and maximum ID must be chosen so that there is at least one ID that will never be used.");
        }
        // Do IDs wrap around? If yes, we have to consider this when we check
        // whether an ID is within the limits.
        this.wrapAround = (maximumId <= minimumId);
        if (blockFromReuseTime < 0) {
            throw new IllegalArgumentException(
                    "The block-from-reuse time must not be negative.");
        }
        this.blockFromReuseTime = blockFromReuseTime;
        this.earlyReuseOnExhaustion = earlyReuseOnExhaustion;
        // The first ID that we start with is the minimum ID.
        this.nextId = new AtomicInteger(minimumId);
        this.reusableIds = new SimpleConcurrentQueue<ReusableId>();
    }

    /**
     * Acquires an ID that is currently not in use. The ID is marked as used so
     * that subsequent calls to this method will not return the ID. It will only
     * be used again after it has been released using the
     * {@link #releaseId(int)} method.
     * 
     * @return unique ID. This ID should be returned to the pool by calling
     *         {@link #releaseId(int)} when it is not needed any longer.
     * @throws IllegalStateException
     *             if this pool is exhausted and no unused ID can be assigned at
     *             the moment.
     */
    public int acquireId() {
        return acquireIdInternal(false);
    }

    /**
     * Acquires an unused ID, preferring IDs that have been released over
     * completely new IDs. If <code>ignoreReuseTime</code> is <code>false</code>
     * and no reusable ID could be found, this method will delegate to
     * {@link #acquireNewId()} in order to acquire a completely new ID. If
     * <code>ignoreReuseTime</code> is <code>true</code> and no reusable ID is
     * found, this method will fail with a {@link IllegalStateException}.
     * 
     * @param ignoreReuseTime
     *            if <code>true</code> even IDs that have been released but that
     *            are still blocked from reuse are considered.
     * @return ID that is currently not in use.
     * @throws IllegalStateException
     *             if no ID can be acquired.
     */
    private int acquireIdInternal(boolean ignoreReuseTime) {
        // We may need the current time several times. In order to reduce the
        // number of system calls, we get it once and reuse it.
        long currentTime = System.currentTimeMillis();
        // Look for an ID that we can reuse.
        ReusableId nextReusableId;
        while ((nextReusableId = reusableIds.peek()) != null) {
            if (nextReusableId.reuseAfter <= currentTime || ignoreReuseTime) {
                // We found a usable ID, now we try to remove it from the queue.
                if (reusableIds.removeIfMatch(nextReusableId)) {
                    // We could remove the ID from the queue. This means that we
                    // can use it.
                    return nextReusableId.id;
                }
                // We could not remove the ID from the queue. This means that a
                // different thread got it. We try again with the next ID:
            } else {
                // The oldest reusable ID is still not old enough. This means
                // that we cannot reuse an ID but have to use a new one.
                break;
            }
        }
        // Either reusableIds is empty or too little time has passed since the
        // oldest ID on the queue has been released. This means that we have to
        // acquire a new ID. If ignoreReuseTime is true, this method has already
        // been called a second time because there are no new IDs available. In
        // this case, we fail right away in order to avoid an infinite loop.
        if (ignoreReuseTime) {
            throw new IllegalStateException(
                    "No ID could be acquired because all possible IDs are currently in use.");
        }
        Integer newId = acquireNewId();
        if (newId != null) {
            return newId;
        } else {
            if (this.earlyReuseOnExhaustion) {
                return acquireIdInternal(true);
            } else {
                throw new IllegalStateException(
                        "No ID could be acquired because all possible IDs are currently in use or still blocked from reuse.");
            }
        }
    }

    /**
     * Acquires a new ID. The ID returned has never been used before. This is
     * ensured by incrementing a counter that keeps the next yet unused ID. When
     * all new IDs have been exhausted (the counter has reached the configured
     * limit), <code>null</code> is returned.
     * 
     * @return new ID or <code>null</code> if the counter for new IDs has been
     *         exhausted.
     */
    private Integer acquireNewId() {
        int currentNextId = this.nextId.get();
        if (!isWithinLimits(currentNextId)) {
            // The next ID counter is exhausted so we cannot acquire a new ID.
            return null;
        }
        // We get the next next ID by incrementing the current next ID. If the
        // current next ID is the maximum Integer value, we wrap around and
        // start using negative values.
        int nextNextId = (currentNextId == Integer.MAX_VALUE) ? Integer.MIN_VALUE
                : (currentNextId + 1);
        // If we can successfully replace the current next ID with the
        // calculated next ID, we can use the current next ID.
        if (this.nextId.compareAndSet(currentNextId, nextNextId)) {
            return currentNextId;
        } else {
            // Another thread was quicker, so we simply retry.
            return acquireNewId();
        }
    }

    /**
     * Checks whether an ID is within the configured limits.
     * 
     * @param id
     *            ID to be checked.
     * @return <code>true</code> if <code>id</code> is within the limits (is a
     *         valid ID), <code>false</code> otherwise.
     */
    private boolean isWithinLimits(int id) {
        if (wrapAround) {
            // If the minimum ID is greater than the maximum ID, an ID is valid
            // if it is above the minimum or below the maximum.
            return id >= minimumId || id <= maximumId;
        } else {
            // If the minimum ID is less than the maximum ID, an ID is valid if
            // it is between the minimum and maximum ID.
            return id >= minimumId && id <= maximumId;
        }
    }

    /**
     * Releases an ID. Releasing an ID puts it back into the pool of available
     * IDs so that it can be returned by a subsequent call to
     * {@link #acquireId()}. If a non-zero block-from-reuse time has been set,
     * the released ID will not be returned by {@link #acquireId()} before this
     * time has passed.
     * 
     * @param id
     *            ID that is not used any longer and should be put back into the
     *            pool of available IDs.
     */
    public void releaseId(int id) {
        ReusableId reusableId = new ReusableId();
        reusableId.id = id;
        reusableId.reuseAfter = System.currentTimeMillis() + blockFromReuseTime;
        this.reusableIds.add(reusableId);
    }

}
