/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.NoSuchElementException;
import java.util.WeakHashMap;

/**
 * <p>
 * Hash set that holds its elements through weak references and has a
 * predictable iteration order.
 * </p>
 * 
 * <p>
 * Internally, this set wraps its elements in {@link WeakReference}s and stores
 * them inside a {@link HashMap}. Unlike a {@link WeakHashMap}, this set uses
 * object identity and not equality when comparing two objects. All elements
 * also form a linked list, so the iterator returns the elements in the order in
 * which they were added. This set does not support <code>null</code> elements.
 * Trying to add a <code>null</code> element causes a
 * {@link NullPointerException} .
 * </p>
 * 
 * <p>
 * Elements in this set are hold through {@link WeakReference}s, so the fact
 * that an object has been added to this set will not keep it from being garbage
 * collected. As the time of garbage collection is unpredictable, this set might
 * change unexpectedly. For example:
 * </p>
 * 
 * <ul>
 * <li>Two subsequent calls to {@link #size()} may return two different numbers
 * where the second number is smaller than the first one, without explicitly
 * removing any elements from the set.</li>
 * <li>{@link #isEmpty()} might return <code>false</code> and then
 * <code>true</code> even though the set has not been modified in between.</li>
 * <li>When iterating over the set several times, subsequent iterations might
 * yield less elements than earlier iterations.</li>
 * </ul>
 * 
 * <p>
 * However, because this set compares elements based on object identity, not
 * equality, some operations are guaranteed to be consistent. For example:
 * </p>
 * 
 * <ul>
 * <li>Several subsequent calls to {@link #contains(Object)} will always return
 * the same result.</li>
 * <li>After adding an object, calls to {@link #contains(Object)} will always
 * return <code>true</code> until the object is removed explicitly.</li>
 * </ul>
 * 
 * <p>
 * Iterators returned by the {@link #iterator()} method have some weak
 * consistency guarantees:
 * </p>
 * 
 * <ul>
 * <li>If {@link Iterator#hasNext() hasNext()} returned <code>true</code>, the
 * {@link Iterator#next() next()} method will always return a non-null element
 * and will never throw a {@link NoSuchElementException}.</li>
 * <li>Even though entries might suddenly disappear due to garbage collection,
 * the iterator will never throw a {@link ConcurrentModificationException}.
 * However, the iterator will throw such an exception if (after creating the
 * iterator) this set is modified explicitly through other means than the
 * iterator's {@link Iterator#remove() remove()} method.</li>
 * </ul>
 * 
 * <p>
 * The performance of this set should be roughly equivalent to the performance
 * of a {@link LinkedHashSet}. However, occasionally stale elements (elements
 * that have been found by the garbage collector to be only weakly reachable)
 * are removed, so an operation which theoretically is O(1) is actually O(n)
 * where n is the number of entries that have been selected by the garbage
 * collector since the last removal of stale elements. Stale elements are
 * removed each time one of this set's methods is called.
 * </p>
 * 
 * <p>
 * This set is not safe for use by multiple threads. In this case, access to
 * this set (and its iterators) needs to be synchronized, e.g. by wrapping it
 * with {@link Collections#synchronizedSet(java.util.Set)}. Unlike most
 * collection implementations, this set is not even safe for read-only use by
 * multiple threads because a read operation can trigger the removal of stale
 * elements.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <E>
 *            the type of elements maintained by this set.
 */
public class LinkedWeakIdentityHashSet<E> extends AbstractSet<E> {

    /**
     * Weak reference holding the actual element. By deriving from
     * {@link WeakReference}, we ensure that the element may be garbage
     * collected. We do not use a standard {@link WeakReference} because we have
     * to override the {@link #equals(Object)} and {@link #hashCode()} methods
     * and need two fields because we also use this class as a linked-list node.
     * 
     * @author Sebastian Marsching
     *
     * @param <E>
     *            the type of elements maintained by this set
     */
    private static class WeakIdentityReference<E> extends WeakReference<E> {

        private int hashCode;
        private WeakIdentityReference<E> next = null;
        private WeakIdentityReference<E> previous = null;

        private WeakIdentityReference(E referent, ReferenceQueue<? super E> q) {
            super(referent, q);
            // We use the identity hash code because we want to compare elements
            // based on object identity. We store the hash code because later,
            // when this reference has been cleared, we cannot get the hash code
            // from the referent any longer.
            this.hashCode = System.identityHashCode(referent);
        }

        @Override
        public int hashCode() {
            // We always return the stored hash code. This ensures that the hash
            // code never changes, even when this reference has been cleared.
            return hashCode;
        }

        @Override
        public boolean equals(Object obj) {
            // We do not allow adding null elements and catch null objects in
            // the contains(...) and remove(...) methods of the set, so this
            // should never happen. However, it is better to be on the safe
            // side.
            if (obj == null) {
                return false;
            }
            if (obj instanceof WeakIdentityReference) {
                // This kind of comparison is used when adding elements and when
                // removing the because they have been garbage collected.
                WeakIdentityReference<?> other = (WeakIdentityReference<?>) obj;
                E referent = this.get();
                if (referent == null) {
                    // If this reference has already been cleared and the other
                    // one has also been cleared, we cannot know for sure
                    // whether they originally referred to the same object.
                    // Therefore, we only have to choices: Either we risk false
                    // positives or false negatives. If we compared the hash
                    // codes, we would risk false positives because two hash
                    // codes might match by accident. We risk false negatives by
                    // comparing the references themselves. Only if the other
                    // object actually is the same reference, we treat it as
                    // equal. This means that two references that have both been
                    // equal because they pointed to the same object are not
                    // equal any longer once they have been cleared. This kind
                    // of comparison should never happen, because we only allow
                    // references to distinct objects to be stored in our set
                    // and for a newly added element the reference should never
                    // have been cleared because there still are strong
                    // references to it.
                    return this == other;
                } else {
                    // We compare the two referents using the equals operator
                    // because we want to compare based on object identity.
                    return referent == other.get();
                }
            } else if (obj instanceof StrongIdentityReference) {
                // When trying to remove an object or checking whether an object
                // is an element of the set, we use a StrongIdentityReference
                // instead of a WeakIdentityReference because it is lighter and
                // there is no risk of comparing null values. Therefore, we also
                // also have to implement the equality operator for this kind of
                // object. We do not have to worry about null values because as
                // long as a strong reference refers to the object, the weak
                // reference cannot have been cleared.
                StrongIdentityReference<?> other = (StrongIdentityReference<?>) obj;
                return this.get() == other.get();
            } else {
                return false;
            }
        }

    }

    /**
     * Strong reference to an object that uses object identity instead of
     * equality when comparing two references.
     * 
     * @author Sebastian Marsching
     *
     * @param <E>
     *            the type of elements maintained by this set
     */
    private static class StrongIdentityReference<E> {

        private E referent;

        private StrongIdentityReference(E referent) {
            assert (referent != null);
            this.referent = referent;
        }

        public E get() {
            return referent;
        }

        @Override
        public int hashCode() {
            return System.identityHashCode(referent);
        }

        @Override
        public boolean equals(Object obj) {
            // We do not allow adding null elements and catch null objects in
            // the contains(...) and remove(...) methods of the set, so this
            // should never happen. However, it is better to be on the safe
            // side.
            if (obj == null) {
                return false;
            }
            // We support comparisons with WeakIdentityReference and
            // StrongIdentityReference objects. This way, we can use strong
            // references when checking for an element (or removing it) and weak
            // references when adding an element.
            if (obj instanceof WeakIdentityReference) {
                WeakIdentityReference<?> other = (WeakIdentityReference<?>) obj;
                return this.referent == other.get();
            } else if (obj instanceof StrongIdentityReference) {
                StrongIdentityReference<?> other = (StrongIdentityReference<?>) obj;
                return this.referent == other.referent;
            } else {
                return false;
            }
        }

    }

    private WeakIdentityReference<E> first = null;
    private WeakIdentityReference<E> last = null;
    private HashMap<WeakIdentityReference<E>, Object> map;
    private transient int modCount = 0;
    private ReferenceQueue<E> queue = new ReferenceQueue<E>();

    /**
     * Creates an empty set with the default initial capacity and load factor of
     * the underlying {@link HashMap} (typically 16 and 0.75).
     */
    public LinkedWeakIdentityHashSet() {
        map = new HashMap<LinkedWeakIdentityHashSet.WeakIdentityReference<E>, Object>();
    }

    /**
     * Creates an empty set with the specified initial capacity and the default
     * load factor of the underlying {@link HashMap} (typically 0.75).
     * 
     * @param initialCapacity
     *            the initial capacity of the hash table.
     * @throws IllegalArgumentException
     *             if the initial capacity is less than zero.
     */
    public LinkedWeakIdentityHashSet(int initialCapacity) {
        map = new HashMap<LinkedWeakIdentityHashSet.WeakIdentityReference<E>, Object>(
                initialCapacity);
    }

    /**
     * Creates an empty set with the specified initial capacity and load factor.
     * 
     * @param initialCapacity
     *            the initial capacity of the hash table.
     * @param loadFactor
     *            the load factor of the hash table.
     * @throws IllegalArgumentException
     *             if the initial capacity is less than zero, or if the load
     *             factor is not positive.
     */
    public LinkedWeakIdentityHashSet(int initialCapacity, float loadFactor) {
        map = new HashMap<LinkedWeakIdentityHashSet.WeakIdentityReference<E>, Object>(
                initialCapacity, loadFactor);
    }

    /**
     * Constructs a new set containing the elements in the specified collection.
     * The underlying {@link HashMap} has a load factor of 0.75 and an initial
     * capacity sufficient to hold all elements from the collection.
     * 
     * @param c
     *            the collection whose elements are to be placed into this set.
     * @throws NullPointerException
     *             if the specified collection is <code>null</code> or contains
     *             <code>null</code> elements.
     */
    public LinkedWeakIdentityHashSet(Collection<? extends E> c) {
        this(Math.max((int) (c.size() / 0.75f) + 1, 16), 0.75f);
        addAll(c);
    }

    private void processReferenceQueue() {
        Reference<? extends E> reference;
        while ((reference = queue.poll()) != null) {
            // By synchronizing on the queue here, the set is safe for read-only
            // use by multiple threads if a concurrently running contains(...)
            // operation can handle the inconsistent state caused by the
            // remove(...) operation. This is true for the HashMap
            // implementation in the Oracle JDK. However, a different platform
            // might use a different implementation were the contains(...)
            // operation fails in case of a concurrent remove(...) operation.
            // Therefore, we do not promise that this class is safe for
            // read-only use by multiple threads.
            synchronized (queue) {
                removeInternal(reference);
            }
        }
    }

    private boolean removeInternal(Object o) {
        if (o == null) {
            return false;
        }
        Object key;
        if (o instanceof WeakIdentityReference) {
            key = o;
        } else {
            // For finding an element, we use a StrongIdentityReference instead
            // of a WeakIdentityReference. This object is lighter and it keeps a
            // strong reference to the element, so there is no risk that the
            // element we are looking for might get garbage collected while
            // looking for it.
            key = new StrongIdentityReference<Object>(o);
        }
        @SuppressWarnings("unchecked")
        WeakIdentityReference<E> entry = (WeakIdentityReference<E>) map
                .remove(key);
        if (entry != null) {
            modCount++;
            WeakIdentityReference<E> previous = entry.previous;
            WeakIdentityReference<E> next = entry.next;
            if (previous != null) {
                previous.next = next;
            } else {
                first = next;
            }
            if (next != null) {
                next.previous = previous;
            } else {
                last = previous;
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    public boolean add(E e) {
        processReferenceQueue();
        if (e == null) {
            throw new NullPointerException();
        }
        WeakIdentityReference<E> entry = new WeakIdentityReference<E>(e, queue);
        boolean modified = (map.put(entry, entry) == null);
        if (modified) {
            modCount++;
            if (last == null) {
                first = last = entry;
            } else {
                last.next = entry;
                entry.previous = last;
                last = entry;
            }
        }
        return modified;
    }

    @Override
    public void clear() {
        map.clear();
        first = last = null;
        modCount++;
        while (queue.poll() != null)
            ;
    }

    @Override
    public boolean contains(Object o) {
        processReferenceQueue();
        if (o == null) {
            return false;
        }
        // For finding an element, we use an IdentityWrapper instead of a
        // WeakIdentityReference. This object is lighter and it keeps a strong
        // reference to the element, so there is no risk that the element we are
        // looking for might get garbage collected while looking for it.
        return map.containsKey(new StrongIdentityReference<Object>(o));
    }

    @Override
    public Iterator<E> iterator() {
        processReferenceQueue();
        return new Iterator<E>() {

            private WeakIdentityReference<E> current = null;
            private int modCountAtStart = modCount;
            private WeakIdentityReference<E> next = null;
            private WeakIdentityReference<E> nextNext = first;
            private E nextValue;

            @Override
            public boolean hasNext() {
                if (modCountAtStart != modCount) {
                    throw new ConcurrentModificationException();
                }
                while (this.nextNext != null && this.nextValue == null) {
                    E nextValue = this.nextNext.get();
                    if (nextValue != null) {
                        this.next = this.nextNext;
                        this.nextValue = nextValue;
                    } else {
                        // The same considerations as for the
                        // processReferenceQueue method apply here.
                        synchronized (queue) {
                            removeInternal(this.nextNext);
                        }
                    }
                    this.nextNext = this.nextNext.next;
                }
                return (nextValue != null);
            }

            @Override
            public E next() {
                if (!hasNext()) {
                    throw new NoSuchElementException();
                }
                E value = nextValue;
                current = next;
                nextValue = null;
                next = null;
                return value;
            }

            @Override
            public void remove() {
                if (current == null) {
                    throw new IllegalStateException();
                }
                if (modCountAtStart != modCount) {
                    throw new ConcurrentModificationException();
                }
                removeInternal(current);
                modCount++;
                modCountAtStart = modCount;
                current = null;
            }

        };
    }

    @Override
    public boolean remove(Object o) {
        if (o == null) {
            return false;
        }
        processReferenceQueue();
        return removeInternal(o);
    }

    @Override
    public int size() {
        processReferenceQueue();
        return map.size();
    }

}
