/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

/**
 * <p>
 * Channel Access value that represent a type with alarm information and
 * possibly additional information (<code>DBR_STS_*</code>,
 * <code>DBR_TIME_*</code>, <code>DBR_GR_*</code>, or <code>DBR_CTRL_*</code>).
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @param <ElementType>
 *            type of the actual value elements. For primitives, this is the
 *            boxed variant.
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessAlarmValue<ElementType> extends
        ChannelAccessGettableValue<ElementType> {

    /**
     * Returns the alarm status of the value. The alarm status provides
     * information about the cause of an alarm (e.g. value too low or too high,
     * invalid value, etc.).
     * 
     * @return alarm status of the value (never <code>null</code>).
     */
    ChannelAccessAlarmStatus getAlarmStatus();

    /**
     * Set the alarm status of the value. The alarm status provides information
     * about the cause of an alarm (e.g. value too low or too high, invalid
     * value, etc.).
     * 
     * @param alarmStatus
     *            new alarm status of the value (never <code>null</code>).
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus);

    /**
     * Returns the alarm severity of the value. The alarm severity signals how
     * severe an alarm situation is.
     * 
     * @return alarm severity of the value (never <code>null</code>).
     */
    ChannelAccessAlarmSeverity getAlarmSeverity();

    /**
     * Sets the alarm severity of the value. The alarm severity signals how
     * severe an alarm situation is.
     * 
     * @param alarmSeverity
     *            new alarm severity of the value (never <code>null</code>).
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity);

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessAlarmValue<ElementType> asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessAlarmValue<ElementType> clone();

}
