/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.charset.Charset;
import java.util.Collection;
import java.util.Collections;

import org.apache.commons.lang3.ArrayUtils;

import com.aquenos.epics.jackie.common.value.internal.ChannelAccessAcknowledgeAlarmImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessAlarmAcknowledgementStatusImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessCharBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessClassNameImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessConfigureAcknowledgeTransientAlarmsImpl;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessDoubleBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessEnumBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessFloatBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessLongBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessShortBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessStringBase;
import com.aquenos.epics.jackie.common.value.internal.ChannelAccessTimeStampImpl;

/**
 * <p>
 * Factory for values representing Channel Access types. This factory provides
 * the methods through which instances of {@link ChannelAccessValue} should be
 * created. Using those methods ensures that the implementations used are
 * compatible with {@link ChannelAccessValueCodec}.
 * </p>
 * 
 * <p>
 * If values should be created from their serialized representation, the methods
 * provided by {@link ChannelAccessValueCodec} should be used instead.
 * </p>
 * 
 * @see ChannelAccessValue
 * @see ChannelAccessValueCodec
 * @see ChannelAccessValueType
 * @author Sebastian Marsching
 */
public final class ChannelAccessValueFactory {

    /**
     * Private default constructor. This class provides static methods and
     * should never be instantiated.
     */
    private ChannelAccessValueFactory() {
    }

    /**
     * <p>
     * Creates a time-stamp <code>seconds</code> seconds and
     * <code>nanoseconds</code> nanoseconds after EPICS epoch. The EPICS epoch
     * is January 1st, 1990, 00:00:00 UTC. EPICS uses unsigned integers, so a
     * negative value for <code>seconds</code> actually indicates a value
     * greater than the {@link Integer#MAX_VALUE maximum positive integer value}
     * .
     * </p>
     * 
     * <p>
     * The <code>nanoseconds</code> part is restricted to the range of 0 to
     * 999999999. Values outside this range are silently confined to this range.
     * </p>
     * 
     * @param seconds
     *            number of seconds since epoch (interpreted as an unsigned
     *            32-bit integer).
     * @param nanoseconds
     *            number of nanoseconds to be added for higher precision.
     * @return time-stamp representing the specified point in time (never
     *         <code>null</code>).
     */
    public static ChannelAccessTimeStamp createTimeStamp(int seconds,
            int nanoseconds) {
        return new ChannelAccessTimeStampImpl(seconds, nanoseconds);
    };

    /**
     * Creates a Channel Access value of the specified type. For most types, the
     * newly created value has no elements. However, the
     * {@link ChannelAccessValueType#DBR_PUT_ACKT DBR_PUT_ACKT} and
     * {@link ChannelAccessValueType#DBR_PUT_ACKS DBR_PUT_ACKS} types are
     * created with on value element (with a value zero).
     * 
     * @param valueType
     *            type of the Channel Access value that should be created.
     * @param charset
     *            charset encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the specified type.
     */
    public static ChannelAccessValue<?> createValue(
            ChannelAccessValueType valueType, Charset charset) {
        switch (valueType) {
        case DBR_STRING:
            return createString(Collections.<String> emptyList(), charset);
        case DBR_SHORT:
            return createShort(ArrayUtils.EMPTY_SHORT_ARRAY);
        case DBR_FLOAT:
            return createFloat(ArrayUtils.EMPTY_FLOAT_ARRAY);
        case DBR_ENUM:
            return createEnum(ArrayUtils.EMPTY_SHORT_ARRAY);
        case DBR_CHAR:
            return createChar(ArrayUtils.EMPTY_BYTE_ARRAY);
        case DBR_LONG:
            return createLong(ArrayUtils.EMPTY_INT_ARRAY);
        case DBR_DOUBLE:
            return createDouble(ArrayUtils.EMPTY_DOUBLE_ARRAY);
        case DBR_STS_STRING:
            return createAlarmString(Collections.<String> emptyList(),
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, charset);
        case DBR_STS_SHORT:
            return createAlarmShort(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STS_FLOAT:
            return createAlarmFloat(ArrayUtils.EMPTY_FLOAT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STS_ENUM:
            return createAlarmEnum(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STS_CHAR:
            return createAlarmChar(ArrayUtils.EMPTY_BYTE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STS_LONG:
            return createAlarmLong(ArrayUtils.EMPTY_INT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STS_DOUBLE:
            return createAlarmDouble(ArrayUtils.EMPTY_DOUBLE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_TIME_STRING:
            return createTimeString(Collections.<String> emptyList(),
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0, charset);
        case DBR_TIME_SHORT:
            return createTimeShort(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_TIME_FLOAT:
            return createTimeFloat(ArrayUtils.EMPTY_FLOAT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_TIME_ENUM:
            return createTimeEnum(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_TIME_CHAR:
            return createTimeChar(ArrayUtils.EMPTY_BYTE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_TIME_LONG:
            return createTimeLong(ArrayUtils.EMPTY_INT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_TIME_DOUBLE:
            return createTimeDouble(ArrayUtils.EMPTY_DOUBLE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, 0, 0);
        case DBR_GR_STRING:
            return createGraphicsString(Collections.<String> emptyList(),
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, charset);
        case DBR_GR_SHORT:
            return createGraphicsShort(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", (short) 0,
                    (short) 0, (short) 0, (short) 0, (short) 0, (short) 0,
                    charset);
        case DBR_GR_FLOAT:
            return createGraphicsFloat(ArrayUtils.EMPTY_FLOAT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, (short) 0, "", 0f, 0f,
                    0f, 0f, 0f, 0f, charset);
        case DBR_GR_ENUM:
            return createGraphicsEnum(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM,
                    Collections.<String> emptyList(), charset);
        case DBR_GR_CHAR:
            return createGraphicsChar(ArrayUtils.EMPTY_BYTE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", (byte) 0,
                    (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, charset);
        case DBR_GR_LONG:
            return createGraphicsLong(ArrayUtils.EMPTY_INT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", 0, 0, 0, 0, 0, 0,
                    charset);
        case DBR_GR_DOUBLE:
            return createGraphicsDouble(ArrayUtils.EMPTY_DOUBLE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, (short) 0, "", 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, charset);
        case DBR_CTRL_STRING:
            return createControlsString(Collections.<String> emptyList(),
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, charset);
        case DBR_CTRL_SHORT:
            return createControlsShort(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", (short) 0,
                    (short) 0, (short) 0, (short) 0, (short) 0, (short) 0,
                    (short) 0, (short) 0, charset);
        case DBR_CTRL_FLOAT:
            return createControlsFloat(ArrayUtils.EMPTY_FLOAT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, (short) 0, "", 0f, 0f,
                    0f, 0f, 0f, 0f, 0f, 0f, charset);
        case DBR_CTRL_ENUM:
            return createControlsEnum(ArrayUtils.EMPTY_SHORT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM,
                    Collections.<String> emptyList(), charset);
        case DBR_CTRL_CHAR:
            return createControlsChar(ArrayUtils.EMPTY_BYTE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", (byte) 0,
                    (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0, (byte) 0,
                    (byte) 0, charset);
        case DBR_CTRL_LONG:
            return createControlsLong(ArrayUtils.EMPTY_INT_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, "", 0, 0, 0, 0, 0, 0,
                    0, 0, charset);
        case DBR_CTRL_DOUBLE:
            return createControlsDouble(ArrayUtils.EMPTY_DOUBLE_ARRAY,
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, (short) 0, "", 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, charset);
        case DBR_PUT_ACKT:
            return createConfigureAcknowledgeTransientAlarms(false);
        case DBR_PUT_ACKS:
            return createAcknowledgeAlarm(ChannelAccessAlarmSeverity.NO_ALARM);
        case DBR_STSACK_STRING:
            return createAlarmAcknowledgementStatus(
                    Collections.<String> emptyList(),
                    ChannelAccessAlarmStatus.NO_ALARM,
                    ChannelAccessAlarmSeverity.NO_ALARM, false,
                    ChannelAccessAlarmSeverity.NO_ALARM, charset);
        case DBR_CLASS_NAME:
            return createClassName(Collections.<String> emptyList(), charset);
        default:
            throw new RuntimeException("Unhandled type " + valueType);
        }
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STRING
     * DBR_STRING}. The collection containing the value elements is copied, so
     * subsequent changes to it will not affect this value and vice versa. The
     * <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STRING DBR_STRING} type.
     */
    public static ChannelAccessSimpleOnlyString createString(
            Collection<String> value, Charset charset) {
        return new ChannelAccessStringBase.ChannelAccessStringImpl(value,
                charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_STRING
     * DBR_STS_STRING}. The collection containing the value elements is copied,
     * so subsequent changes to it will not affect this value and vice versa.
     * The <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING}
     *         type.
     */
    public static ChannelAccessAlarmOnlyString createAlarmString(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, Charset charset) {
        return new ChannelAccessStringBase.ChannelAccessAlarmStringImpl(
                ChannelAccessValueType.DBR_STS_STRING, value, alarmStatus,
                alarmSeverity, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_STRING
     * DBR_TIME_STRING}. The collection containing the value elements is copied,
     * so subsequent changes to it will not affect this value and vice versa.
     * The <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING}
     *         type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeString createTimeString(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds, Charset charset) {
        return new ChannelAccessStringBase.ChannelAccessTimeStringImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_STRING
     * DBR_GR_STRING}. The collection containing the value elements is copied,
     * so subsequent changes to it will not affect this value and vice versa.
     * The <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} type.
     */
    public static ChannelAccessAlarmOnlyString createGraphicsString(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, Charset charset) {
        return new ChannelAccessStringBase.ChannelAccessAlarmStringImpl(
                ChannelAccessValueType.DBR_GR_STRING, value, alarmStatus,
                alarmSeverity, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_STRING
     * DBR_CTRL_STRING}. The collection containing the value elements is copied,
     * so subsequent changes to it will not affect this value and vice versa.
     * The <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING}
     *         type.
     */
    public static ChannelAccessAlarmOnlyString createControlsString(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, Charset charset) {
        return new ChannelAccessStringBase.ChannelAccessAlarmStringImpl(
                ChannelAccessValueType.DBR_CTRL_STRING, value, alarmStatus,
                alarmSeverity, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_SHORT
     * DBR_SHORT}. The array is not copied, so subsequent changes in the array
     * will affect this value and vice versa.
     * 
     * @param value
     *            the shorts that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} type.
     */
    public static ChannelAccessSimpleOnlyShort createShort(short[] value) {
        return new ChannelAccessShortBase.ChannelAccessShortImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_SHORT
     * DBR_STS_SHORT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the shorts that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} type.
     */
    public static ChannelAccessAlarmOnlyShort createAlarmShort(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessShortBase.ChannelAccessAlarmShortImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_SHORT
     * DBR_TIME_SHORT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the shorts that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT}
     *         type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeShort createTimeShort(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessShortBase.ChannelAccessTimeShortImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_SHORT
     * DBR_GR_SHORT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the shorts that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} type.
     */
    public static ChannelAccessGraphicsOnlyShort createGraphicsShort(
            short[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            short upperDisplayLimit, short lowerDisplayLimit,
            short upperAlarmLimit, short upperWarningLimit,
            short lowerWarningLimit, short lowerAlarmLimit, Charset charset) {
        return new ChannelAccessShortBase.ChannelAccessGraphicsShortImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_SHORT
     * DBR_CTRL_SHORT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the shorts that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the valu (never
     *            <code>null</code>)e.
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperControlLimit
     *            upper control limit. The upper control limit is the upper
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is greater than this limit. A value of zero typically
     *            indicates that the limit has not been configured and thus
     *            should not be used.
     * @param lowerControlLimit
     *            lower control limit. The lower control limit is the lower
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is less than this limit. A value of zero typically indicates
     *            that the limit has not been configured and thus should not be
     *            used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT}
     *         type.
     */
    public static ChannelAccessControlsShort createControlsShort(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            short upperDisplayLimit, short lowerDisplayLimit,
            short upperAlarmLimit, short upperWarningLimit,
            short lowerWarningLimit, short lowerAlarmLimit,
            short upperControlLimit, short lowerControlLimit, Charset charset) {
        return new ChannelAccessShortBase.ChannelAccessControlsShortImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                upperControlLimit, lowerControlLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_FLOAT
     * DBR_FLOAT}. The array is not copied, so subsequent changes in the array
     * will affect this value and vice versa.
     * 
     * @param value
     *            the floats that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} type.
     */
    public static ChannelAccessSimpleOnlyFloat createFloat(float[] value) {
        return new ChannelAccessFloatBase.ChannelAccessFloatImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_FLOAT
     * DBR_STS_FLOAT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the floats that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} type.
     */
    public static ChannelAccessAlarmOnlyFloat createAlarmFloat(float[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessFloatBase.ChannelAccessAlarmFloatImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_FLOAT
     * DBR_TIME_FLOAT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the floats that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT}
     *         type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeFloat createTimeFloat(float[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessFloatBase.ChannelAccessTimeFloatImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_FLOAT
     * DBR_GR_FLOAT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the floats that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param precision
     *            display precision. The display precision is the number of
     *            digits of the fractional part that should be displayed (number
     *            of digits after the decimal separator). A value of zero
     *            indicates that no display precision has been specified
     *            explicitly and some user-defined default should be used.
     *            Channel Access uses an unsigned 16-bit integer for this
     *            property, so negative values are possible and actually
     *            represent numbers greater than 32767. For practical purposes
     *            however, such values do not make any sense.
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} type.
     */
    public static ChannelAccessGraphicsOnlyFloat createGraphicsFloat(
            float[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, short precision,
            String engineeringUnits, float upperDisplayLimit,
            float lowerDisplayLimit, float upperAlarmLimit,
            float upperWarningLimit, float lowerWarningLimit,
            float lowerAlarmLimit, Charset charset) {
        return new ChannelAccessFloatBase.ChannelAccessGraphicsFloatImpl(value,
                alarmStatus, alarmSeverity, precision, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_FLOAT
     * DBR_CTRL_FLOAT}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the floats that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param precision
     *            display precision. The display precision is the number of
     *            digits of the fractional part that should be displayed (number
     *            of digits after the decimal separator). A value of zero
     *            indicates that no display precision has been specified
     *            explicitly and some user-defined default should be used.
     *            Channel Access uses an unsigned 16-bit integer for this
     *            property, so negative values are possible and actually
     *            represent numbers greater than 32767. For practical purposes
     *            however, such values do not make any sense.
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperControlLimit
     *            upper control limit. The upper control limit is the upper
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is greater than this limit. A value of zero typically
     *            indicates that the limit has not been configured and thus
     *            should not be used.
     * @param lowerControlLimit
     *            lower control limit. The lower control limit is the lower
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is less than this limit. A value of zero typically indicates
     *            that the limit has not been configured and thus should not be
     *            used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT}
     *         type.
     */
    public static ChannelAccessControlsFloat createControlsFloat(float[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, short precision,
            String engineeringUnits, float upperDisplayLimit,
            float lowerDisplayLimit, float upperAlarmLimit,
            float upperWarningLimit, float lowerWarningLimit,
            float lowerAlarmLimit, float upperControlLimit,
            float lowerControlLimit, Charset charset) {
        return new ChannelAccessFloatBase.ChannelAccessControlsFloatImpl(value,
                alarmStatus, alarmSeverity, precision, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                upperControlLimit, lowerControlLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM}.
     * The array is not copied, so subsequent changes in the array will affect
     * this value and vice versa.
     * 
     * @param value
     *            the enums that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the {@link ChannelAccessValueType#DBR_ENUM
     *         DBR_ENUM} type.
     */
    public static ChannelAccessSimpleOnlyEnum createEnum(short[] value) {
        return new ChannelAccessEnumBase.ChannelAccessEnumImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_ENUM
     * DBR_STS_ENUM}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the enums that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} type.
     */
    public static ChannelAccessAlarmOnlyEnum createAlarmEnum(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessEnumBase.ChannelAccessAlarmEnumImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_ENUM
     * DBR_TIME_ENUM}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the enums that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeEnum createTimeEnum(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessEnumBase.ChannelAccessTimeEnumImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_ENUM
     * DBR_GR_ENUM}. The <code>value</code> array is not copied, so subsequent
     * changes in the array will affect this value and vice versa. The
     * <code>labels</code> array, however, is copied.
     * 
     * @param value
     *            the enums that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param labels
     *            enum state labels (never <code>null</code>). Must not contain
     *            <code>null</code> elements. These are display labels that
     *            serve as the textual representation of up to 16 states. Labels
     *            that are too long are truncated automatically when converting
     *            them to raw strings. However, this does not affect the labels
     *            stored in the Java strings. The collection is copied so that
     *            subsequent modifications do not affect the value created.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} type.
     * @throws IllegalArgumentException
     *             if <code>labels</code> has more than 16 elements.
     */
    public static ChannelAccessGraphicsEnum createGraphicsEnum(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity,
            Collection<String> labels, Charset charset) {
        return new ChannelAccessEnumBase.ChannelAccessGraphicsEnumImpl(
                ChannelAccessValueType.DBR_GR_ENUM, value, alarmStatus,
                alarmSeverity, labels, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_ENUM
     * DBR_CTRL_ENUM}. The <code>value</code> array is not copied, so subsequent
     * changes in the array will affect this value and vice versa. The
     * <code>labels</code> array, however, is copied.
     * 
     * @param value
     *            the enums that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param labels
     *            enum state labels (never <code>null</code>). These are display
     *            labels that serve as the textual representation of up to 16
     *            states. Labels that are too long are truncated automatically
     *            when converting them to raw strings. However, this does not
     *            affect the labels stored in the Java strings. The collection
     *            is copied so that subsequent modifications do not affect the
     *            value created.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT}
     *         type.
     * @throws IllegalArgumentException
     *             if <code>labels</code> has more than 16 elements.
     */
    public static ChannelAccessGraphicsEnum createControlsEnum(short[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity,
            Collection<String> labels, Charset charset) {
        return new ChannelAccessEnumBase.ChannelAccessGraphicsEnumImpl(
                ChannelAccessValueType.DBR_CTRL_ENUM, value, alarmStatus,
                alarmSeverity, labels, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR}.
     * The array is not copied, so subsequent changes in the array will affect
     * this value and vice versa.
     * 
     * @param value
     *            the bytes that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the {@link ChannelAccessValueType#DBR_CHAR
     *         DBR_CHAR} type.
     */
    public static ChannelAccessSimpleOnlyChar createChar(byte[] value) {
        return new ChannelAccessCharBase.ChannelAccessCharImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_CHAR
     * DBR_STS_CHAR}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the bytes that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} type.
     */
    public static ChannelAccessAlarmOnlyChar createAlarmChar(byte[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessCharBase.ChannelAccessAlarmCharImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_CHAR
     * DBR_TIME_CHAR}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the bytes that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeChar createTimeChar(byte[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessCharBase.ChannelAccessTimeCharImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_CHAR
     * DBR_GR_CHAR}. The array is not copied, so subsequent changes in the array
     * will affect this value and vice versa.
     * 
     * @param value
     *            the bytes that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} type.
     */
    public static ChannelAccessGraphicsOnlyChar createGraphicsChar(
            byte[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            byte upperDisplayLimit, byte lowerDisplayLimit,
            byte upperAlarmLimit, byte upperWarningLimit,
            byte lowerWarningLimit, byte lowerAlarmLimit, Charset charset) {
        return new ChannelAccessCharBase.ChannelAccessGraphicsCharImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_CHAR
     * DBR_CTRL_CHAR}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the bytes that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperControlLimit
     *            upper control limit. The upper control limit is the upper
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is greater than this limit. A value of zero typically
     *            indicates that the limit has not been configured and thus
     *            should not be used.
     * @param lowerControlLimit
     *            lower control limit. The lower control limit is the lower
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is less than this limit. A value of zero typically indicates
     *            that the limit has not been configured and thus should not be
     *            used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} type.
     */
    public static ChannelAccessControlsChar createControlsChar(byte[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            byte upperDisplayLimit, byte lowerDisplayLimit,
            byte upperAlarmLimit, byte upperWarningLimit,
            byte lowerWarningLimit, byte lowerAlarmLimit,
            byte upperControlLimit, byte lowerControlLimit, Charset charset) {
        return new ChannelAccessCharBase.ChannelAccessControlsCharImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                upperControlLimit, lowerControlLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_LONG DBR_LONG}.
     * The array is not copied, so subsequent changes in the array will affect
     * this value and vice versa.
     * 
     * @param value
     *            the ints that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the {@link ChannelAccessValueType#DBR_LONG
     *         DBR_LONG} type.
     */
    public static ChannelAccessSimpleOnlyLong createLong(int[] value) {
        return new ChannelAccessLongBase.ChannelAccessLongImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_LONG
     * DBR_STS_LONG}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the ints that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} type.
     */
    public static ChannelAccessAlarmOnlyLong createAlarmLong(int[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessLongBase.ChannelAccessAlarmLongImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_LONG
     * DBR_TIME_LONG}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the ints that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeLong createTimeLong(int[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessLongBase.ChannelAccessTimeLongImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_LONG
     * DBR_GR_LONG}. The array is not copied, so subsequent changes in the array
     * will affect this value and vice versa.
     * 
     * @param value
     *            the ints that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} type.
     */
    public static ChannelAccessGraphicsOnlyLong createGraphicsLong(int[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            int upperDisplayLimit, int lowerDisplayLimit, int upperAlarmLimit,
            int upperWarningLimit, int lowerWarningLimit, int lowerAlarmLimit,
            Charset charset) {
        return new ChannelAccessLongBase.ChannelAccessGraphicsLongImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_LONG
     * DBR_CTRL_LONG}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the ints that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperControlLimit
     *            upper control limit. The upper control limit is the upper
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is greater than this limit. A value of zero typically
     *            indicates that the limit has not been configured and thus
     *            should not be used.
     * @param lowerControlLimit
     *            lower control limit. The lower control limit is the lower
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is less than this limit. A value of zero typically indicates
     *            that the limit has not been configured and thus should not be
     *            used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} type.
     */
    public static ChannelAccessControlsLong createControlsLong(int[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, String engineeringUnits,
            int upperDisplayLimit, int lowerDisplayLimit, int upperAlarmLimit,
            int upperWarningLimit, int lowerWarningLimit, int lowerAlarmLimit,
            int upperControlLimit, int lowerControlLimit, Charset charset) {
        return new ChannelAccessLongBase.ChannelAccessControlsLongImpl(value,
                alarmStatus, alarmSeverity, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                upperControlLimit, lowerControlLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_DOUBLE
     * DBR_DOUBLE}. The array is not copied, so subsequent changes in the array
     * will affect this value and vice versa.
     * 
     * @param value
     *            the doubles that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} type.
     */
    public static ChannelAccessSimpleOnlyDouble createDouble(double[] value) {
        return new ChannelAccessDoubleBase.ChannelAccessDoubleImpl(value);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STS_DOUBLE
     * DBR_STS_DOUBLE}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the doubles that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE}
     *         type.
     */
    public static ChannelAccessAlarmOnlyDouble createAlarmDouble(
            double[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity) {
        return new ChannelAccessDoubleBase.ChannelAccessAlarmDoubleImpl(value,
                alarmStatus, alarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_TIME_DOUBLE
     * DBR_TIME_DOUBLE}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the doubles that should be stored in the value (never
     *            <code>null</code>).
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param timeStampSeconds
     *            seconds part of the time-stamp. These are the seconds since
     *            EPICS epoch (January 1st, 1990, 00:00:00 UTC). EPICS uses
     *            unsigned integers, so a negative value for
     *            <code>timeStampseconds</code> actually indicates a value
     *            greater than the {@link Integer#MAX_VALUE maximum positive
     *            integer value}.
     * @param timeStampNanoSeconds
     *            nanoseconds part of the time-stamp. The nanoseconds are
     *            specified so that the time-stamp can have a higher precision.
     *            This value must be between 0 and 999999999.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE}
     *         type.
     * @throws IllegalArgumentException
     *             if <code>timeStampNanoSeconds</code> is out of range.
     */
    public static ChannelAccessTimeDouble createTimeDouble(double[] value,
            ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, int timeStampSeconds,
            int timeStampNanoSeconds) {
        return new ChannelAccessDoubleBase.ChannelAccessTimeDoubleImpl(value,
                alarmStatus, alarmSeverity, timeStampSeconds,
                timeStampNanoSeconds);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_GR_DOUBLE
     * DBR_GR_DOUBLE}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the doubles that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param precision
     *            display precision. The display precision is the number of
     *            digits of the fractional part that should be displayed (number
     *            of digits after the decimal separator). A value of zero
     *            indicates that no display precision has been specified
     *            explicitly and some user-defined default should be used.
     *            Channel Access uses an unsigned 16-bit integer for this
     *            property, so negative values are possible and actually
     *            represent numbers greater than 32767. For practical purposes
     *            however, such values do not make any sense.
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} type.
     */
    public static ChannelAccessGraphicsOnlyDouble createGraphicsDouble(
            double[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, short precision,
            String engineeringUnits, double upperDisplayLimit,
            double lowerDisplayLimit, double upperAlarmLimit,
            double upperWarningLimit, double lowerWarningLimit,
            double lowerAlarmLimit, Charset charset) {
        return new ChannelAccessDoubleBase.ChannelAccessGraphicsDoubleImpl(
                value, alarmStatus, alarmSeverity, precision, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CTRL_DOUBLE
     * DBR_CTRL_DOUBLE}. The array is not copied, so subsequent changes in the
     * array will affect this value and vice versa.
     * 
     * @param value
     *            the doubles that should be stored in the value (never
     *            <code>null</code>). The array is not copied.
     * @param alarmStatus
     *            the alarm status associated with the value (never
     *            <code>null</code>).
     * @param alarmSeverity
     *            the alarm severity associated with the value (never
     *            <code>null</code>).
     * @param precision
     *            display precision. The display precision is the number of
     *            digits of the fractional part that should be displayed (number
     *            of digits after the decimal separator). A value of zero
     *            indicates that no display precision has been specified
     *            explicitly and some user-defined default should be used.
     *            Channel Access uses an unsigned 16-bit integer for this
     *            property, so negative values are possible and actually
     *            represent numbers greater than 32767. For practical purposes
     *            however, such values do not make any sense.
     * @param engineeringUnits
     *            engineerings units for the value (never <code>null</code>).
     * @param upperDisplayLimit
     *            upper display limit. The upper display limit is the limit that
     *            the value does typically not exceed and thus is suitable for
     *            choosing the layout of graphic displays (e.g. gauges). A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerDisplayLimit
     *            lower display limit. The lower display limit is the limit that
     *            the value does typically not undershoot and thus is suitable
     *            for choosing the layout of graphic displays (e.g. gauges). A
     *            value of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperAlarmLimit
     *            upper alarm limit. The upper alarm limit is the limit that,
     *            when exceeded, typically raises a major alarm. A value of zero
     *            typically indicates that the limit has not been configured and
     *            thus should not be used.
     * @param upperWarningLimit
     *            upper warning limit. The upper warning limit is the limit
     *            that, when exceeded, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerWarningLimit
     *            lower warning limit. The lower warning limit is the limit
     *            that, when undershot, typically raises a minor alarm. A value
     *            of zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param lowerAlarmLimit
     *            lower alarm limit. The lower alarm limit is the limit that,
     *            when undershot, typically raises a major alarm. A value of
     *            zero typically indicates that the limit has not been
     *            configured and thus should not be used.
     * @param upperControlLimit
     *            upper control limit. The upper control limit is the upper
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is greater than this limit. A value of zero typically
     *            indicates that the limit has not been configured and thus
     *            should not be used.
     * @param lowerControlLimit
     *            lower control limit. The lower control limit is the lower
     *            limit that the server enforces on values. Typically, this
     *            means that the process variable cannot be set to a value that
     *            is less than this limit. A value of zero typically indicates
     *            that the limit has not been configured and thus should not be
     *            used.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE}
     *         type.
     */
    public static ChannelAccessControlsDouble createControlsDouble(
            double[] value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity alarmSeverity, short precision,
            String engineeringUnits, double upperDisplayLimit,
            double lowerDisplayLimit, double upperAlarmLimit,
            double upperWarningLimit, double lowerWarningLimit,
            double lowerAlarmLimit, double upperControlLimit,
            double lowerControlLimit, Charset charset) {
        return new ChannelAccessDoubleBase.ChannelAccessControlsDoubleImpl(
                value, alarmStatus, alarmSeverity, precision, engineeringUnits,
                upperDisplayLimit, lowerDisplayLimit, upperAlarmLimit,
                upperWarningLimit, lowerWarningLimit, lowerAlarmLimit,
                upperControlLimit, lowerControlLimit, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_PUT_ACKT
     * DBR_PUT_ACKT}.
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            <code>true</code> if transient alarms should have to be
     *            acknowledged explicitly, <code>false</code> if they should be
     *            automatically acknowledged when they disappear.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_PUT_ACKT DBR_PUT_ACKT} type.
     */
    public static ChannelAccessConfigureAcknowledgeTransientAlarms createConfigureAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms) {
        return new ChannelAccessConfigureAcknowledgeTransientAlarmsImpl(
                explicitlyAcknowledgeTransientAlarms);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_PUT_ACKS
     * DBR_PUT_ACKS}.
     * 
     * @param acknowledgeAlarmSeverity
     *            highest alarm severity that should be acknowledged (never
     *            <code>null</code>). This means that the alarm will be
     *            acknowledged if the severity of the unacknowledged alarm is
     *            less than or equal to the specified severity.
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_PUT_ACKS DBR_PUT_ACKS} type.
     */
    public static ChannelAccessAcknowledgeAlarm createAcknowledgeAlarm(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity) {
        return new ChannelAccessAcknowledgeAlarmImpl(acknowledgeAlarmSeverity);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_STSACK_STRING
     * DBR_STSACK_STRING}. The collection containing the value elements is
     * copied, so subsequent changes to it will not affect this value and vice
     * versa. The <code>value</code> collection must not contain any
     * <code>null</code> elements. If it does, this causes a
     * {@link NullPointerException}.
     * 
     * @param value
     *            strings representing the channel's current value (never
     *            <code>null</code>). The collection is copied and must not
     *            contain <code>null</code> elements.
     * @param alarmStatus
     *            current alarm status (never <code>null</code>).
     * @param currentAlarmSeverity
     *            current alarm severity (never <code>null</code>).
     * @param explicitlyAcknowledgeTransientAlarms
     *            configuration for transient alarms. Do transient alarms have
     *            to be acknowledged explicitly (<code>true</code>) or are they
     *            automatically acknowledged when they disappear (
     *            <code>false</code>)?
     * @param unackowledgedAlarmSeverity
     *            severity of the highest unacknowledged alarm (never
     *            <code>null</code>).
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_STSACK_STRING
     *         DBR_STSACK_STRING} type.
     */
    public static ChannelAccessAlarmAcknowledgementStatus createAlarmAcknowledgementStatus(
            Collection<String> value, ChannelAccessAlarmStatus alarmStatus,
            ChannelAccessAlarmSeverity currentAlarmSeverity,
            boolean explicitlyAcknowledgeTransientAlarms,
            ChannelAccessAlarmSeverity unackowledgedAlarmSeverity,
            Charset charset) {
        return new ChannelAccessAlarmAcknowledgementStatusImpl(value,
                alarmStatus, currentAlarmSeverity,
                explicitlyAcknowledgeTransientAlarms,
                unackowledgedAlarmSeverity, charset);
    }

    /**
     * Creates a value of type {@link ChannelAccessValueType#DBR_CLASS_NAME
     * DBR_CLASS_NAME}. The collection containing the value elements is copied,
     * so subsequent changes to it will not affect this value and vice versa.
     * The <code>value</code> collection must not contain any <code>null</code>
     * elements. If it does, this causes a {@link NullPointerException}.
     * 
     * @param value
     *            the strings that should be stored in the value (never
     *            <code>null</code>). The array is copied and must not contain
     *            <code>null</code> elements.
     * @param charset
     *            encoding used for encoding and decoding strings (never
     *            <code>null</code>).
     * @return new value instance of the
     *         {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME}
     *         type.
     */
    public static ChannelAccessClassName createClassName(
            Collection<String> value, Charset charset) {
        return new ChannelAccessClassNameImpl(value, charset);
    }

}
