/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;

import java.nio.BufferOverflowException;
import java.nio.ByteBuffer;

import org.junit.Test;

public class ByteBufferByteSinkTest {

    @Test
    public void singleByteBuffer() {
        ByteBuffer buffer = ByteBuffer.allocate(100);
        ByteBufferByteSink byteSink = new ByteBufferByteSink()
                .addBuffer(buffer);
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 0L);
        byteSink.putByte((byte) 2);
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 1L);
        byteSink.putInt(35);
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 5L);
        byteSink.putDouble(5.3);
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 13L);
        short[] shortArray = new short[] { 5, -7, 12341, 41, 17 };
        byteSink.putShortArray(shortArray);
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 23L);
        ByteBuffer[] fullBuffers = byteSink.getWrittenData();
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 0L);
        assertEquals(1, fullBuffers.length);
        buffer = fullBuffers[0];
        assertEquals(2, buffer.get());
        assertEquals(35, buffer.getInt());
        assertEquals(5.3, buffer.getDouble(), 0.0);
        short[] shortArray2 = new short[shortArray.length];
        for (int i = 0; i < shortArray2.length; i++) {
            shortArray2[i] = buffer.getShort();
        }
        assertArrayEquals(shortArray, shortArray2);
        assertEquals(0, buffer.remaining());
    }

    @Test
    public void emptyPut() {
        ByteBuffer buffer = ByteBuffer.allocate(1);
        ByteBufferByteSink byteSink = new ByteBufferByteSink()
                .addBuffer(buffer);
        byteSink.putDoubleArray(new double[0]);
        assertEquals(0, buffer.position());
        assertEquals(byteSink.getWrittenDataSizeInBytes(), 0L);
    }

    @Test
    public void multipleByteBuffers() {
        ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffers(new ByteBuffer[] { ByteBuffer.allocate(15),
                ByteBuffer.allocate(8), ByteBuffer.allocate(0),
                ByteBuffer.allocate(27), ByteBuffer.allocate(88) });

        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        byte[] byteArray = new byte[] { 5, 3, -7, 2, 8 };
        sink.putByteArray(byteArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 5L);
        double[] doubleArray = new double[] { 3.6, -4.7e15 };
        sink.putDoubleArray(doubleArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
        float[] floatArray = new float[] { 4.3F, 8.9F, -17.3F };
        sink.putFloatArray(floatArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 33L);
        int[] intArray = new int[] { -1124, 7241, 715, 71241, 8 };
        sink.putIntArray(intArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 53L);
        short[] shortArray = new short[] { 18, 12, -414, 724, 715, 8121, -24215 };
        sink.putShortArray(shortArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 67L);
        long[] longArray = new long[] { 5L, 0L, 9999999999L, -18L,
                Long.MIN_VALUE, 78L };
        sink.putLongArray(longArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 115L);
        ByteBuffer[] fullBuffers = sink.getWrittenData();
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        ByteBufferByteSource source = new ByteBufferByteSource()
                .appendBuffers(fullBuffers);
        assertArrayEquals(byteArray, source.getByteArray(5));
        assertArrayEquals(doubleArray, source.getDoubleArray(2), 0.0);
        assertArrayEquals(floatArray, source.getFloatArray(3), 0.0F);
        assertArrayEquals(intArray, source.getIntArray(5));
        assertArrayEquals(shortArray, source.getShortArray(7));
        assertArrayEquals(longArray, source.getLongArray(6));
        for (ByteBuffer buffer : fullBuffers) {
            assertFalse(buffer.hasRemaining());
        }
    }

    @Test
    public void partialArrayPut() {
        int[] intArray = new int[] { 0, 1, 2, 3, 4, 5, 6, 7 };
        ByteBufferByteSink sink = new ByteBufferByteSink().addBuffer(ByteBuffer
                .allocate(24));
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        sink.putIntArray(intArray, 1, 6);
        assertEquals(sink.getWrittenDataSizeInBytes(), 24L);
        ByteBuffer buffer = sink.getWrittenData()[0];
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        assertEquals(1, buffer.getInt());
        assertEquals(2, buffer.getInt());
        assertEquals(3, buffer.getInt());
        assertEquals(4, buffer.getInt());
        assertEquals(5, buffer.getInt());
        assertEquals(6, buffer.getInt());
        assertEquals(0, buffer.remaining());
    }

    @Test(expected = BufferOverflowException.class)
    public void writeOverLimit() {
        ByteBufferByteSink sink = new ByteBufferByteSink().addBuffer(ByteBuffer
                .allocate(7));
        sink.putLong(5L);
    }

    @Test
    public void consistentPut() {
        ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffers(new ByteBuffer[] { ByteBuffer.allocate(13),
                ByteBuffer.allocate(8), ByteBuffer.allocate(27) });

        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        byte[] byteArray = new byte[] { 5, 3, -7, 2, 8 };
        sink.putByteArray(byteArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 5L);
        double[] doubleArray = new double[] { 3.6, -4.7e15 };
        sink.putDoubleArray(doubleArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
        float[] floatArray = new float[] { 4.3F, 8.9F, -17.3F };
        sink.putFloatArray(floatArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 33L);
        int[] intArray = new int[] { -1124, 7241, 715, 71241, 8 };
        try {
            sink.putIntArray(intArray);
        } catch (BufferOverflowException e) {
            // We expect this exception, the interesting question is, whether
            // the data has been written in a consistent way. This means that
            // the everything but the integers should have been written.
        }
        assertEquals(sink.getWrittenDataSizeInBytes(), 33L);
        ByteBuffer[] fullBuffers = sink.getWrittenData();
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        ByteBufferByteSource source = new ByteBufferByteSource()
                .appendBuffers(fullBuffers);
        assertArrayEquals(byteArray, source.getByteArray(5));
        assertArrayEquals(doubleArray, source.getDoubleArray(2), 0.0);
        assertArrayEquals(floatArray, source.getFloatArray(3), 0.0F);
        for (ByteBuffer buffer : fullBuffers) {
            assertFalse(buffer.hasRemaining());
        }
    }

    @Test
    public void atomicPutNoError() {
        final ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffers(new ByteBuffer[] { ByteBuffer.allocate(15),
                ByteBuffer.allocate(8), ByteBuffer.allocate(0),
                ByteBuffer.allocate(27), ByteBuffer.allocate(88) });

        final byte[] byteArray = new byte[] { 5, 3, -7, 2, 8 };
        final double[] doubleArray = new double[] { 3.6, -4.7e15 };
        final float[] floatArray = new float[] { 4.3F, 8.9F, -17.3F };
        final int[] intArray = new int[] { -1124, 7241, 715, 71241, 8 };
        final short[] shortArray = new short[] { 18, 12, -414, 724, 715, 8121,
                -24215 };
        final long[] longArray = new long[] { 5L, 0L, 9999999999L, -18L,
                Long.MIN_VALUE, 78L };
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        sink.putByteArray(byteArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 5L);
        sink.putDoubleArray(doubleArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
        sink.atomicPut(new ByteSink.AtomicPutOperation<Void>() {
            @Override
            public Void put() {
                assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
                sink.putFloatArray(floatArray);
                assertEquals(sink.getWrittenDataSizeInBytes(), 33L);
                sink.putIntArray(intArray);
                assertEquals(sink.getWrittenDataSizeInBytes(), 53L);
                sink.putShortArray(shortArray);
                assertEquals(sink.getWrittenDataSizeInBytes(), 67L);
                sink.putLongArray(longArray);
                assertEquals(sink.getWrittenDataSizeInBytes(), 115L);
                return null;
            }
        });
        assertEquals(sink.getWrittenDataSizeInBytes(), 115L);
        ByteBuffer[] fullBuffers = sink.getWrittenData();
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        ByteBufferByteSource source = new ByteBufferByteSource()
                .appendBuffers(fullBuffers);
        assertArrayEquals(byteArray, source.getByteArray(5));
        assertArrayEquals(doubleArray, source.getDoubleArray(2), 0.0);
        assertArrayEquals(floatArray, source.getFloatArray(3), 0.0F);
        assertArrayEquals(intArray, source.getIntArray(5));
        assertArrayEquals(shortArray, source.getShortArray(7));
        assertArrayEquals(longArray, source.getLongArray(6));
        for (ByteBuffer buffer : fullBuffers) {
            assertFalse(buffer.hasRemaining());
        }
    }

    @Test
    public void atomicPutWithError() {
        final ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffers(new ByteBuffer[] { ByteBuffer.allocate(13),
                ByteBuffer.allocate(8), ByteBuffer.allocate(27) });

        final byte[] byteArray = new byte[] { 5, 3, -7, 2, 8 };
        final double[] doubleArray = new double[] { 3.6, -4.7e15 };
        final float[] floatArray = new float[] { 4.3F, 8.9F, -17.3F };
        final int[] intArray = new int[] { -1124, 7241, 715, 71241, 8 };
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        sink.putByteArray(byteArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 5L);
        sink.putDoubleArray(doubleArray);
        assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
        try {
            sink.atomicPut(new ByteSink.AtomicPutOperation<Void>() {
                @Override
                public Void put() {
                    assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
                    sink.putFloatArray(floatArray);
                    assertEquals(sink.getWrittenDataSizeInBytes(), 33L);
                    sink.putIntArray(intArray);
                    return null;
                }
            });
        } catch (BufferOverflowException e) {
            // We expect this exception, the interesting question is, whether
            // the data has been written in a consistent way. This means that
            // neither the floats nor the integers should have been written.
        }
        assertEquals(sink.getWrittenDataSizeInBytes(), 21L);
        ByteBuffer[] fullBuffers = sink.getWrittenData();
        assertEquals(sink.getWrittenDataSizeInBytes(), 0L);
        ByteBufferByteSource source = new ByteBufferByteSource()
                .appendBuffers(fullBuffers);
        assertArrayEquals(byteArray, source.getByteArray(5));
        assertArrayEquals(doubleArray, source.getDoubleArray(2), 0.0);
        for (ByteBuffer buffer : fullBuffers) {
            assertFalse(buffer.hasRemaining());
        }
    }

    /**
     * Test the <code>addBufferFirst</code> method.
     */
    @Test
    public void addBufferFirst() {
        ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffer(ByteBuffer.allocate(16));
        sink.addBuffer(ByteBuffer.allocate(16));
        assertEquals(0L, sink.getWrittenDataSizeInBytes());
        sink.putFloat(4.3f);
        assertEquals(4L, sink.getWrittenDataSizeInBytes());
        ByteBuffer[] buffers = sink.getWrittenData();
        assertEquals(1, buffers.length);
        ByteBuffer buffer = buffers[0];
        assertEquals(4, buffer.remaining());
        assertEquals(0L, sink.getWrittenDataSizeInBytes());
        buffer.position(buffer.limit());
        buffer.limit(buffer.capacity());
        sink.addBufferFirst(buffer);
        assertEquals(4L, sink.getWrittenDataSizeInBytes());
        sink.putShort((short) 5);
        assertEquals(6L, sink.getWrittenDataSizeInBytes());
        buffers = sink.getWrittenData();
        assertEquals(1, buffers.length);
        buffer = buffers[0];
        assertEquals(6, buffer.remaining());
    }

    /**
     * Test the <code>addBufferFirst</code> method. This method should fail with
     * an {@link IllegalStateException}, because it is not allowed to prepend a
     * buffer to a byte sink that already contains data.
     */
    @Test(expected = IllegalStateException.class)
    public void addBufferFirstFailIfSinkNotEmpty() {
        ByteBufferByteSink sink = new ByteBufferByteSink();
        sink.addBuffer(ByteBuffer.allocate(16));
        sink.addBuffer(ByteBuffer.allocate(16));
        assertEquals(0L, sink.getWrittenDataSizeInBytes());
        sink.putFloat(4.3f);
        assertEquals(4L, sink.getWrittenDataSizeInBytes());
        ByteBuffer[] buffers = sink.getWrittenData();
        assertEquals(1, buffers.length);
        ByteBuffer buffer = buffers[0];
        assertEquals(4, buffer.remaining());
        assertEquals(0L, sink.getWrittenDataSizeInBytes());
        buffer.position(buffer.limit());
        buffer.limit(buffer.capacity());
        sink.putInt(2);
        assertEquals(4, sink.getWrittenDataSizeInBytes());
        // Should throw.
        sink.addBufferFirst(buffer);
    }

}
