/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import org.apache.commons.lang3.tuple.Pair;
import org.junit.Test;

/**
 * Tests for {@link BasicIntegerIdPool}.
 * 
 * @author Sebastian Marsching
 */
public class BasicIntegerIdPoolTest {

    /**
     * Tests whether a released ID is reused after a configured time. This test
     * is done with a single thread because it is hard to test this feature when
     * multiple tests run in parallel.
     */
    @Test
    public void reuseAfterDelay() throws InterruptedException {
        BasicIntegerIdPool idPool;
        idPool = new BasicIntegerIdPool(0, Integer.MAX_VALUE, 500L, false);
        assertEquals(0, idPool.acquireId());
        assertEquals(1, idPool.acquireId());
        idPool.releaseId(0);
        assertEquals(2, idPool.acquireId());
        // We sleep for 0.8 seconds, so that the released ID should now be
        // reusable.
        Thread.sleep(800L);
        assertEquals(0, idPool.acquireId());
    }

    /**
     * Tests that construction fails when using an invalid ID range.
     */
    @Test(expected = IllegalArgumentException.class)
    public void invalidLimits1() {
        new BasicIntegerIdPool(Integer.MIN_VALUE, Integer.MAX_VALUE, 0L, false);
    }

    /**
     * Tests that construction fails when using an invalid ID range.
     */
    @Test(expected = IllegalArgumentException.class)
    public void invalidLimits2() {
        new BasicIntegerIdPool(0, 0, 0L, false);
    }

    /**
     * Tests that construction fails when using an invalid ID range.
     */
    @Test(expected = IllegalArgumentException.class)
    public void invalidLimits3() {
        new BasicIntegerIdPool(0, -1, 0L, false);
    }

    /**
     * Tests that construction fails when using an invalid block-from-reuse
     * time.
     */
    @Test(expected = IllegalArgumentException.class)
    public void invalidBlockFromReuseTime() {
        new BasicIntegerIdPool(0, Integer.MAX_VALUE, -1L, false);
    }

    /**
     * Run test with several concurrent threads. This test tries to expose
     * concurrency issues by requesting IDs and releasing them from multiple
     * concurrent threads.
     */
    @Test
    public void parallelTests() throws InterruptedException, ExecutionException {
        // We use the same number of threads and the same number of IDs per
        // thread for all tests. Those numbers are designed to result in a
        // reasonable running time for each test.
        int numberOfThreads = 10;
        int numberOfIdsPerThread = 100000;
        int totalNumberOfIds = numberOfThreads * numberOfIdsPerThread;
        Pair<HashSet<Integer>, HashSet<Integer>> result;
        HashSet<Integer> acquiredIds;
        HashSet<Integer> releasedIds;
        // Test positive IDs only, no block from reuse.
        result = runParallelTest(numberOfThreads, numberOfIdsPerThread, 0,
                Integer.MAX_VALUE, 0L, false);
        acquiredIds = result.getLeft();
        releasedIds = result.getRight();
        // We expect that all IDs from 0 to totalNumberOfIds - 1 have been
        // acquired, because all released IDs should have been reused.
        for (int id = 0; id < totalNumberOfIds - 1; id++) {
            assertTrue(acquiredIds.contains(id));
        }
        // Test negative IDs only, block reuse for 50 milliseconds.
        result = runParallelTest(numberOfThreads, numberOfIdsPerThread,
                Integer.MIN_VALUE, -1, 50L, false);
        acquiredIds = result.getLeft();
        releasedIds = result.getRight();
        for (int id : acquiredIds) {
            assertTrue(id < 0);
        }
        // Test negative and positive IDs, blocking reuse for a very long time.
        // This means that effectively, only new IDs are going to be assigned.
        result = runParallelTest(numberOfThreads, numberOfIdsPerThread,
                Integer.MIN_VALUE + 1, Integer.MAX_VALUE, 3600000L, false);
        acquiredIds = result.getLeft();
        releasedIds = result.getRight();
        for (int id = Integer.MIN_VALUE + 1, numberOfAcquiredIds = 0; numberOfAcquiredIds < totalNumberOfIds; id++) {
            boolean acquired = acquiredIds.contains(id);
            boolean released = releasedIds.contains(id);
            if (acquired) {
                numberOfAcquiredIds++;
            }
            // Each ID must either have been acquired or must have been released
            // (but never both).
            assertTrue(acquired || released && !(acquired && released));
        }
        // Test IDs wrapping from positive to negative.
        result = runParallelTest(numberOfThreads, numberOfIdsPerThread,
                Integer.MAX_VALUE - 50000, -1, 0L, false);
        acquiredIds = result.getLeft();
        releasedIds = result.getRight();
        for (int id : acquiredIds) {
            assertTrue(id >= Integer.MAX_VALUE - 50000 || id <= -1);
        }
        // Test a pool that will get exhausted because we block reuse for a very
        // long time. However, the task should still succeed, because we allow
        // early reuse on exhaustion.
        result = runParallelTest(numberOfThreads, numberOfIdsPerThread, 0,
                totalNumberOfIds - 1, 3600000L, true);
        acquiredIds = result.getLeft();
        releasedIds = result.getRight();
        assertTrue(acquiredIds.size() == totalNumberOfIds);
        for (int id = 0; id < totalNumberOfIds; id++) {
            // All possible IDs should have been acquired even though they had
            // not been released for reuse yet.
            assertTrue(acquiredIds.contains(id));
        }
    }

    /**
     * Run an exhaustion test with several concurrent threads. This test tries
     * to expose concurrency issues by requesting IDs and releasing them from
     * multiple concurrent threads. The task should fail with an
     * {@link IllegalStateException} because there are not enough IDs in the
     * pool.
     */
    @Test(expected = IllegalStateException.class)
    public void parallelExhaustion() throws InterruptedException,
            ExecutionException {
        int numberOfThreads = 10;
        int numberOfIdsPerThread = 100000;
        int totalNumberOfIds = numberOfThreads * numberOfIdsPerThread;
        // We block IDs from reuse for a sufficiently long time, so that they
        // should not be reused and at some point, ID allocation should fail.
        runParallelTest(numberOfThreads, numberOfIdsPerThread, 0,
                totalNumberOfIds, 3600000L, false);
    }

    /**
     * Run an exhaustion test with several concurrent threads. This test tries
     * to expose concurrency issues by requesting IDs and releasing them from
     * multiple concurrent threads. The task should fail with an
     * {@link IllegalStateException} because there are not enough possible IDs.
     */
    @Test(expected = IllegalStateException.class)
    public void parallelExhaustionWithReuse() throws InterruptedException,
            ExecutionException {
        int numberOfThreads = 10;
        int numberOfIdsPerThread = 100000;
        int totalNumberOfIds = numberOfThreads * numberOfIdsPerThread;
        // We choose the number of possible IDs so that the pool will get
        // exhausted, even after reusing all IDs.
        runParallelTest(numberOfThreads, numberOfIdsPerThread, 0,
                totalNumberOfIds - 10, 3600000L, true);
    }

    /**
     * Threads acquiring IDs from an {@link BasicIntegerIdPool}.
     */
    private class IdRequester implements
            Callable<Pair<LinkedList<Integer>, LinkedList<Integer>>> {

        private BasicIntegerIdPool idPool;
        private int numberOfIds;

        public IdRequester(BasicIntegerIdPool idPool, int numberOfIds) {
            this.idPool = idPool;
            this.numberOfIds = numberOfIds;
        }

        @Override
        public Pair<LinkedList<Integer>, LinkedList<Integer>> call()
                throws Exception {
            LinkedList<Integer> acquiredIds = new LinkedList<Integer>();
            LinkedList<Integer> releasedIds = new LinkedList<Integer>();
            Random random = new Random();
            while (acquiredIds.size() < numberOfIds) {
                acquiredIds.add(idPool.acquireId());
                if (random.nextFloat() < 0.3f) {
                    // In 30 % of the iterations we release one of the IDs from
                    // the start of our list (older IDs).
                    int indexToRelease = random.nextInt(10);
                    if (indexToRelease >= acquiredIds.size()) {
                        indexToRelease = acquiredIds.size() - 1;
                    }
                    int idToRelease = acquiredIds.remove(indexToRelease);
                    idPool.releaseId(idToRelease);
                    releasedIds.add(idToRelease);
                }
            }
            return Pair.of(acquiredIds, releasedIds);
        }

    }

    private Pair<HashSet<Integer>, HashSet<Integer>> runParallelTest(
            int numberOfThreads, int numberOfIdsPerThread, int minimumId,
            int maximumId, long blockFromReuseTime,
            boolean earlyReuseOnExhaustion) throws InterruptedException,
            ExecutionException {
        BasicIntegerIdPool idPool = new BasicIntegerIdPool(minimumId,
                maximumId, blockFromReuseTime, earlyReuseOnExhaustion);
        // Create and start threads.
        // Create and start receiver threads.
        ArrayList<FutureTask<Pair<LinkedList<Integer>, LinkedList<Integer>>>> tasks = new ArrayList<FutureTask<Pair<LinkedList<Integer>, LinkedList<Integer>>>>(
                numberOfThreads);
        for (int i = 0; i < numberOfThreads; i++) {
            FutureTask<Pair<LinkedList<Integer>, LinkedList<Integer>>> task = new FutureTask<Pair<LinkedList<Integer>, LinkedList<Integer>>>(
                    new IdRequester(idPool, numberOfIdsPerThread));
            tasks.add(task);
            new Thread(task).start();
        }
        // Collect IDs acquired and check that they are all unique.
        HashSet<Integer> allAcquiredIds = new HashSet<Integer>(numberOfThreads
                * numberOfIdsPerThread);
        HashSet<Integer> allReleasedIds = new HashSet<Integer>();
        int totalCount = 0;
        for (FutureTask<Pair<LinkedList<Integer>, LinkedList<Integer>>> task : tasks) {
            Pair<LinkedList<Integer>, LinkedList<Integer>> result;
            try {
                result = task.get();
            } catch (ExecutionException e) {
                Throwable cause = e.getCause();
                if (cause != null && cause instanceof RuntimeException) {
                    throw (RuntimeException) cause;
                } else {
                    throw e;
                }
            }
            LinkedList<Integer> acquiredIds = result.getLeft();
            LinkedList<Integer> releasedIds = result.getRight();
            allAcquiredIds.addAll(acquiredIds);
            totalCount += acquiredIds.size();
            allReleasedIds.addAll(releasedIds);
        }
        // Check that we acquired the expected number of IDs in total.
        assertEquals(numberOfThreads * numberOfIdsPerThread, totalCount);
        // Check that all received objects are distinct.
        assertEquals(numberOfThreads * numberOfIdsPerThread,
                allAcquiredIds.size());
        return Pair.of(allAcquiredIds, allReleasedIds);
    }

}
